"""Module inteded to contain some prince-specific data structures."""
import pickle as pickle
import os.path as path
import numpy as np
import scipy.constants as spc
import h5py


from prince_cr.util import convert_to_namedtuple, info
import prince_cr.config as config

#: Dictionary containing particle properties, like mass, charge
#: lifetime or branching ratios
try:
    spec_data = pickle.load(
        open(path.join(config.data_dir, "particle_data.ppo"), "rb"))
except UnicodeDecodeError:
    spec_data = pickle.load(
        open(path.join(config.data_dir, "particle_data.ppo"), "rb"), encoding='latin1')
except FileNotFoundError:
    info(0, 'Warning, particle database "particle_data.ppo" file not found.')


# Default units in Prince are ***cm, s, GeV***
# Define here all constants and unit conversions and use
# throughout the code. Don't write c=2.99.. whatever.
# Write clearly which units a function returns.
# Convert them if not standard unit
# Accept only arguments in the units above

UNITS_AND_CONVERSIONS_DEF = dict(
    c=1e2 * spc.c,
    cm2Mpc=1. / (spc.parsec * spc.mega * 1e2),
    Mpc2cm=spc.mega * spc.parsec * 1e2,
    m_proton=spc.physical_constants['proton mass energy equivalent in MeV'][0]
    * 1e-3,
    m_electron=spc.physical_constants['electron mass energy equivalent in MeV']
    [0] * 1e-3,
    r_electron=spc.physical_constants['classical electron radius'][0] * 1e2,
    fine_structure=spc.fine_structure,
    GeV2erg=1. / 624.15,
    erg2GeV=624.15,
    km2cm=1e5,
    yr2sec=spc.year,
    Gyr2sec=spc.giga * spc.year,
    cm2sec=1e-2 / spc.c,
    sec2cm=spc.c * 1e2)

# This is the immutable unit object to be imported throughout the code
PRINCE_UNITS = convert_to_namedtuple(UNITS_AND_CONVERSIONS_DEF, "PriNCeUnits")


class PrinceDB(object):
    """Provides access to data stored in an HDF5 file.

    The file contains all tables for runnin PriNCe. Currently
    the only still required file is the particle database. The tools
    to generate this database are publicly available in
    `PriNCe-data-utils <https://github.com/joheinze/PriNCe-data-utils>`_.

    """

    def __init__(self):

        info(2, 'Opening HDF5 file', config.db_fname)
        self.prince_db_fname = path.join(config.data_dir, config.db_fname)
        if not path.isfile(self.prince_db_fname):
            raise Exception(
                'Prince DB file {0} not found in "data" directory.'.format(
                    config.db_fname))

        with h5py.File(self.prince_db_fname, 'r') as prince_db:
            self.version = (prince_db.attrs['version'])

    def _check_subgroup_exists(self, subgroup, mname):
        available_models = list(subgroup)
        if mname not in available_models:
            info(0, 'Invalid choice/model', mname)
            info(0, 'Choose from:\n', '\n'.join(available_models))
            raise Exception('Unknown selections.')

    def photo_nuclear_db(self, model_tag):
        info(10, 'Reading photo-nuclear db. tag={0}'.format(model_tag))
        db_entry = {}
        with h5py.File(self.prince_db_fname, 'r') as prince_db:
            self._check_subgroup_exists(prince_db['photo_nuclear'],
                                        model_tag)
            for entry in ['energy_grid', 'fragment_yields', 'inel_mothers',
                          'inelastic_cross_sctions', 'mothers_daughters']:
                info(10, 'Reading entry {0} from db.'.format(entry))
                db_entry[entry] = prince_db['photo_nuclear'][model_tag][entry][:]
        return db_entry

    def photo_meson_db(self, model_tag):
        info(10, 'Reading photo-nuclear db. tag={0}'.format(model_tag))
        db_entry = {}
        with h5py.File(self.prince_db_fname, 'r') as prince_db:
            self._check_subgroup_exists(prince_db['photo_nuclear'],
                                        model_tag)
            for entry in ['energy_grid', 'xbins', 'fragment_yields', 'inel_mothers',
                          'inelastic_cross_sctions', 'mothers_daughters']:
                info(10, 'Reading entry {0} from db.'.format(entry))
                db_entry[entry] = prince_db['photo_nuclear'][model_tag][entry][:]
        return db_entry

    def ebl_spline(self, model_tag, subset='base'):
        from scipy.interpolate import interp2d
        info(10, 'Reading EBL field splines. tag={0}'.format(model_tag))
        with h5py.File(self.prince_db_fname, 'r') as prince_db:
            self._check_subgroup_exists(prince_db['EBL_models'],
                                        model_tag)
            self._check_subgroup_exists(prince_db['EBL_models'][model_tag],
                                        subset)
            spl_gr = prince_db['EBL_models'][model_tag][subset]

            return interp2d(spl_gr['x'], spl_gr['y'], spl_gr['z'],
                            fill_value=0., kind='linear')

#: db_handler is the HDF file interface
db_handler = PrinceDB()

class EnergyGrid(object):
    """Class for constructing a grid for discrete distributions.

    Since we discretize everything in energy, the name seems appropriate.
    All grids are log spaced.

    Args:
        lower (float): log10 of low edge of the lowest bin
        upper (float): log10 of upper edge of the highest bin
        bins_dec (int): bins per decade of energy
    """

    def __init__(self, lower, upper, bins_dec):
        self.bins = np.logspace(lower, upper,
                                int((upper - lower) * bins_dec + 1))
        self.grid = 0.5 * (self.bins[1:] + self.bins[:-1])
        self.widths = self.bins[1:] - self.bins[:-1]
        self.d = self.grid.size
        info(
            5, 'Energy grid initialized {0:3.1e} - {1:3.1e}, {2} bins'.format(
                self.bins[0], self.bins[-1], self.grid.size))


class PrinceSpecies(object):
    """Bundles different particle properties for simplified
    availability of particle properties in :class:`prince.core.PriNCeRun`.

    Args:
      pdgid (int): PDG ID of the particle
      particle_db (object): a dictionary with particle properties
      d (int): dimension of the energy grid
    """
    @staticmethod
    def calc_AZN(nco_id):
        """Returns mass number :math:`A`, charge :math:`Z` and neutron
        number :math:`N` of ``nco_id``."""
        Z, A = 1, 1

        if nco_id >= 100:
            Z = nco_id % 100
            A = (nco_id - Z) // 100
        else:
            Z, A = 0, 0

        return A, Z, A - Z

    def __init__(self, ncoid, princeidx, d):

        info(5, 'Initializing new species', ncoid)

        #: Neucosma ID of particle
        self.ncoid = ncoid
        #: (bool) particle is a hadron (meson or baryon)
        self.is_hadron = False
        #: (bool) particle is a meson
        self.is_meson = False
        #: (bool) particle is a baryon
        self.is_baryon = False
        #: (bool) particle is a lepton
        self.is_lepton = False
        #: (bool) if it's an electromagnetic particle
        self.is_em = False
        #: (bool) particle is a lepton
        self.is_charged = False
        #: (bool) particle is a nucleus
        self.is_nucleus = False
        #: (bool) particle has an energy redistribution
        self.has_redist = False
        #: (bool) particle is stable
        self.is_stable = True
        #: (float) lifetime
        self.lifetime = np.inf
        #: (bool) particle is an alias (PDG ID encodes special scoring behavior)
        self.is_alias = False
        #: (str) species name in string representation
        self.sname = None
        #: decay channels if any
        self.decay_channels = {}
        #: Mass, charge, neutron number
        self.A, self.Z, self.N = 1, None, None
        #: Mass in atomic units or GeV
        self.mass = None

        #: (int) Prince index (in state vector)
        self.princeidx = princeidx

        # (dict) Dimension of energy grids (for idx calculations)
        self.grid_dims = {'default': d}

        # Obtain values for the attributes
        self._init_species()

    def _init_species(self):
        """Fill all class attributes with values from
        :var:`spec_data`, depending on ncoid."""
        ncoid = self.ncoid
        dbentry = spec_data[ncoid]

        if ncoid < 200:
            self.is_nucleus = False
            if ncoid == 0:
                self.is_em = True
            elif ncoid in [100, 101]:
                self.is_hadron = True
                self.is_baryon = True
                self.is_nucleus = True
                self.A, self.Z, self.N = self.calc_AZN(ncoid)
            elif ncoid not in [2, 3, 4, 50]:
                self.is_hadron = True
                self.is_meson = True
            else:
                self.is_lepton = True
                if ncoid in [20, 21]:
                    self.is_em = True
                elif ncoid in [7, 10]:
                    self.is_alias = True
        else:
            self.is_nucleus = True
            self.A, self.Z, self.N = self.calc_AZN(ncoid)

        self.AZN = self.A, self.Z, self.N

        if ncoid <= config.redist_threshold_ID:
            self.has_redist = True

        if "name" not in dbentry:
            info(5, "Name for species", ncoid, "not defined")
            self.sname = "nucleus_{0}".format(ncoid)
        else:
            self.sname = dbentry["name"]

        self.charge = dbentry["charge"]
        self.is_charged = self.charge != 0
        self.is_stable = dbentry["stable"]
        self.lifetime = dbentry["lifetime"]
        self.mass = dbentry["mass"]
        self.decay_channels = dbentry["branchings"]

    @property
    def sl(self):
        """Return the slice for this species on the grid
           can be used as spec[s.sl]

        Returns:
          (slice): a slice object pointing to the species in the state vecgtor
        """
        idx = self.princeidx
        dim = self.grid_dims['default']
        return slice(idx * dim, (idx + 1) * dim)

    def lidx(self, grid_tag='default'):
        """Returns lower index of particle range in state vector.

        Returns:
          (int): lower index in state vector :attr:`PrinceRun.phi`
        """
        return self.princeidx * self.grid_dims[grid_tag]

    def uidx(self, grid_tag='default'):
        """Returns upper index of particle range in state vector.

        Returns:
          (int): upper index in state vector :attr:`PrinceRun.phi`
        """
        return (self.princeidx + 1) * self.grid_dims[grid_tag]

    def lbin(self, grid_tag='default'):
        """Returns lower bin of particle range in state vector.

        Returns:
          (int): lower bin in state vector :attr:`PrinceRun.phi`
        """
        return self.princeidx * (self.grid_dims[grid_tag] + 1)

    def ubin(self, grid_tag='default'):
        """Returns upper bin of particle range in state vector.

        Returns:
          (int): upper bin in state vector :attr:`PrinceRun.phi`
        """
        return (self.princeidx + 1) * (self.grid_dims[grid_tag] + 1)

    def indices(self, grid_tag='default'):
        """Returns a list of all indices in the state vector.

        Returns:
          (numpy.array): array of indices in state vector :attr:`PrinceRun.phi`
        """
        idx = self.princeidx
        dim = self.grid_dims[grid_tag]
        return np.arange(idx * dim, (idx + 1) * dim)


class SpeciesManager(object):
    """Provides a database with particle and species."""

    def __init__(self, ncoid_list, ed):
        # (dict) Dimension of primary grid
        self.grid_dims = {'default': ed}
        # Particle index shortcuts
        #: (dict) Converts Neucosma ID to index in state vector
        self.ncoid2princeidx = {}
        #: (dict) Converts particle name to index in state vector
        self.sname2princeidx = {}
        #: (dict) Converts Neucosma ID to reference of
        # :class:`data.PrinceSpecies`
        self.ncoid2sref = {}
        #: (dict) Converts particle name to reference of
        #:class:`data.PrinceSpecies`
        self.sname2sref = {}
        #: (dict) Converts prince index to reference of
        #:class:`data.PrinceSpecies`
        self.princeidx2sref = {}
        #: (dict) Converts index in state vector to Neucosma ID
        self.princeidx2ncoid = {}
        #: (dict) Converts index in state vector to reference
        # of :class:`data.PrinceSpecies`
        self.princeidx2pname = {}
        #: (int) Total number of species
        self.nspec = 0

        self._gen_species(ncoid_list)
        self._init_species_tables()

    def _gen_species(self, ncoid_list):
        info(4, "Generating list of species.")

        # ncoid_list += spec_data["non_nuclear_species"]

        # Make sure list is unique and sorted
        ncoid_list = sorted(list(set(ncoid_list)))

        self.species_refs = []
        # Define position in state vector (princeidx) by simply
        # incrementing it with the (sorted) list of Neucosma IDs
        for princeidx, ncoid in enumerate(ncoid_list):
            info(
                4, "Appending species {0} at position {1}".format(
                    ncoid, princeidx))
            self.species_refs.append(
                PrinceSpecies(ncoid, princeidx, self.grid_dims['default']))

        self.known_species = [s.ncoid for s in self.species_refs]
        self.redist_species = [
            s.ncoid for s in self.species_refs if s.has_redist
        ]
        self.boost_conserv_species = [
            s.ncoid for s in self.species_refs if not s.has_redist
        ]

    def _init_species_tables(self):
        for s in self.species_refs:
            self.ncoid2princeidx[s.ncoid] = s.princeidx
            self.sname2princeidx[s.sname] = s.princeidx
            self.princeidx2ncoid[s.princeidx] = s.ncoid
            self.princeidx2pname[s.princeidx] = s.sname
            self.ncoid2sref[s.ncoid] = s
            self.princeidx2sref[s.princeidx] = s
            self.sname2sref[s.sname] = s

        self.nspec = len(self.species_refs)

    def add_grid(self, grid_tag, dimension):
        """Defines additional grid dimensions under a certain tag.

        Propagates changes to this variable to all known species.
        """
        info(2, 'New grid_tag', grid_tag, 'with dimension', dimension)
        self.grid_dims[grid_tag] = dimension

        for s in self.species_refs:
            s.grid_dims = self.grid_dims

    def __repr__(self):
        str_out = ""
        ident = 3 * ' '
        for s in self.species_refs:
            str_out += s.sname + '\n' + ident
            str_out += 'NCO id : ' + str(s.ncoid) + '\n' + ident
            str_out += 'PriNCe idx : ' + str(s.princeidx) + '\n\n'

        return str_out



'''
Contains basic functions to handle standard cosmology
'''

import numpy as np

import prince_cr.config as config

def H(z, H0=config.H_0s):
    """Expansion rate of the universe.

    :math:`H(z) = H_0 \\sqrt{\\Omega_m (1 + z)^3 + \\Omega_\\Lambda}`
    Args:
      z (float): redshift

    Returns:
      float: expansion rate in :math:`s^{-1}`
    """

    return H0 * np.sqrt(config.Omega_m * (1 + z)**3 + config.Omega_Lambda)


def star_formation_rate(z, z_inhom=0.):
    """Returns the star formation rate, per comoving volume, evaluated at the specified redshift.

    Ref:
        A.M. Hopkins and J.F. Beacom, Astrophys. J. 651, 142 (2006) [astro-ph/060146]
        P. Baerwald, S. Huemmer, and W. Winter, Astropart. Phys. 35, 508 (2012) [1107.5583]

    Args:
      z (float): redshift
      z_inhom (float): redshift where the universe becomes inhomogenous, return 0. below this values

    Returns:
      float: star formation rate normalized to 1. at y = 0.
    """

    if z < z_inhom:
        return 0.
    elif z <= 0.97:
        return (1. + z)**3.44
    elif 0.97 < z <= 4.48:
        return 10.**1.09 * (1. + z)**-0.26
    else:
        return 10.**6.66 * (1. + z)**-7.8


def grb_rate(z, z_inhom=0.):
    """Returns the rate of Gamma-Ray Burst, per comoving volume, evaluated at the specified redshift.

    Ref:
        TODO: Add reference

    Args:
      z (float): redshift
      z_inhom (float): redshift where the universe becomes inhomogenous, return 0. below this values

    Returns:
      float: GRB rate normalized to 1. at y = 0.
    """
    return (1 + z)**1.4 * star_formation_rate(z, z_inhom=z_inhom)

def grb_rate_wp(z, z_inhom=0.):
    """Returns the rate of Gamma-Ray Burst, per comoving volume, evaluated at the specified redshift.

    Ref:
        Wanderman, Piran: Mon.Not.Roy.Astron.Soc.406:1944-1958,2010

    Args:
      z (float): redshift
      z_inhom (float): redshift where the universe becomes inhomogenous, return 0. below this values

    Returns:
      float: GRB rate normalized to 1. at y = 0.
    """
    if z < z_inhom:
        return 0.
    elif z <= 3:
        return (1 + z)**2.1
    else:
        return (1 + 3)**(2.1 + 1.4) * (1 + z)**-1.4


def agn_rate(z, z_inhom=0.):
    """Returns the rate of Active Galactic Nuclei, per comoving volume, evaluated at the specified redshift.

    Ref:
        TODO: Add reference

    Args:
      z (float): redshift
      z_inhom (float): redshift where the universe becomes inhomogenous, return 0. below this values

    Returns:
      float: AGN rate normalized to 1. at y = 0.
    """
    if z < z_inhom:
        return 0.
    elif z <= 1.7:
        return (1 + z)**5
    elif 1.7 < z <= 2.7:
        return (1 + 1.7)**5
    else:
        return (1 + 1.7)**5 * 10**(2.7 - z)


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    zrange = np.linspace(0, 6, 30)
    sfr = np.vectorize(star_formation_rate)
    fig, ax = plt.subplots(1, 1, figsize=(4, 3))
    ax.plot(zrange, sfr(zrange))
    ax.set_xlabel(r'Redshift $z$')
    ax.set_ylabel(r'Star formation rate Mpc$^{-3}$')

    fig, ax = plt.subplots(1, 1, figsize=(4, 3))
    ax.plot(zrange, H(zrange, H0=config.H_0s))
    ax.set_xlabel(r'Redshift $z$')
    ax.set_ylabel(r'Expansion rate of the universe (km s$^{-1}$ Mpc$^{-1}$')

    plt.show()



"""PriNCe configuration module."""

import os.path as path
import platform
import sys
import warnings

import numpy as np

base_path = path.dirname(path.abspath(__file__))

#: Debug flag for verbose printing, 0 silences PriNCe entirely
debug_level = 1
#: Printout debug info only for functions in this list (just give the name,
#: "get_solution" for instance) Warning, this option slows down initialization
#: by a lot. Use only when needed.
override_debug_fcn = []
#: Override debug printout for debug levels < value for the functions above
override_max_level = 10
#: Print module name in debug output
print_module = False

# =================================================================
# Paths and library locations
# =================================================================

#: Directory where the data files for the calculation are stored
data_dir = path.join(base_path, 'data')

#: PrinceDB file name
db_fname = 'prince_db_05.h5'

#: Model file for redistribution functions (from SOPHIA or similar)
redist_fname = 'sophia_redistribution_logbins.npy'


#=========================================================================
# Physics configuration
#=========================================================================

#: Cosmological parameters

#: Hubble constant
H_0 = 70.5  #km s^-1 Mpc^-1
H_0s = 2.28475e-18  #s^-1

#: Omega_m
Omega_m = 0.27

#: Omega_Lambda
Omega_Lambda = 0.73

#: CMB energy kB*T0 [GeV]
E_CMB = 2.34823e-13  

#===========================================================================
# Grids
#===========================================================================

#: Cosmic ray energy grid (defines system size for solver)
#: Number of bins in multiples of 4 recommended for maximal vectorization
#: efficiency for 256 bit AVX or similar
#: Format (log10(E_min), log10(E_max), nbins/decade of energy)
cosmic_ray_grid = (3, 14, 8)
#: Photon grid of target field, only for calculation of rates
photon_grid = (-15, -6, 8)

#: Scale of the energy grid
#:'E': logarithmic in energy E_i = E_min * (Delta)^i
#:'logE': linear grid in x = log_10(E): x_i = x_min + i * Delta
grid_scale ='E'

#: Order of semi-lagrangian for energy derivative 
semi_lagr_method ='5th_order'

#===========================================================================
# Model options
#===========================================================================

#: Threshold lifetime value for explicit transport of particles of this type. It
#: means that if a particle is unstable with lifetime smaller than this threshold,
#: it will be decayed until all final state particles of this chain are stable.
#: In other words: short intermediate states will be integrated out
tau_dec_threshold = np.inf # All unstable particles decay
# tau_dec_threshold = 0.  # None unstable particles decay
# tau_dec_threshold = 850. # This value is for stable neutrons

#: Particle ID for which redistribution functions are needed to be taken into
#: account. The default value is 101 (proton). All particles with smaller
#: IDs, i.e. neutrinos, pions, muons etc., will have energy redistributions.
#: For larger IDs (nuclei) the boost conservation is employed.
redist_threshold_ID = 101

#: Cut on energy redistribution functions
#: Resitribution below this x value are set to 0.
#: "x_cut" : 0.,
#: "x_cut_proton" : 0.,
x_cut = 1e-4
x_cut_proton = 1e-1

#: cut on photon energy, cross section above y = E_cr e_ph / m_cr does not contribute
y_cut = np.inf

# Build equation system up to a maximal nuclear mass of
max_mass = np.inf

# Include secondaries like photons and neutrinos
secondaries = True
# List of specific particles to ignore
ignore_particles = [20,21] # (we ignore photons and electrons, as their physics is not fully implemented)

#===========================================================================
# Parameters of numerical integration
#===========================================================================

# Update rates at not more frequently than this value in z
update_rates_z_threshold = 0.01

# #Number of MKL threads (for sparse matrix multiplication the performance
# #advantage from using more than a few threads is limited by memory bandwidth)
MKL_threads = 4

# Sparse matrix-vector product from "CUPY"|"MKL"|"scipy"
linear_algebra_backend = "MKL"


# Check for CUPY library for GPU support
try:
    import cupy
    has_cupy = True
    mempool = cupy.get_default_memory_pool()
    mempool.free_all_blocks()
except ModuleNotFoundError:
    print('CUPY not found for GPU support. Degrading to MKL.')
    if linear_algebra_backend == 'cupy':
        linear_algebra_backend = 'MKL'
    has_cupy = False

#: determine shared library extension and MKL path
pf = platform.platform()

if 'Linux' in pf:
    mkl_path = path.join(sys.prefix, 'lib', 'libmkl_rt.so')
elif 'Darwin' in pf:
    mkl_path = path.join(sys.prefix, 'lib', 'libmkl_rt.dylib')
else:
    # Windows case
    mkl_path = path.join(sys.prefix, 'Library', 'bin', 'mkl_rt.dll')

# mkl library handler
mkl = None

# Check if MKL library found
if path.isfile(mkl_path):
    has_mkl = True
else:
    has_mkl = False

def set_mkl_threads(nthreads):
    global mkl, MKL_threads
    from ctypes import cdll, byref, c_int
    mkl = cdll.LoadLibrary(mkl_path)
    # Set number of threads
    MKL_threads = nthreads
    mkl.mkl_set_num_threads(byref(c_int(nthreads)))
    if debug_level >= 5:
        print('MKL threads limited to {0}'.format(nthreads))

if has_mkl:
    set_mkl_threads(MKL_threads)

if not has_mkl and linear_algebra_backend.lower() == 'mkl':
    print('MKL runtime not found. Degrading to scipy.')
    linear_algebra_backend = 'scipy'

def _download_file(url, outfile):
    """Downloads the PriNCe database from github release binaries."""

    from tqdm import tqdm
    import requests
    import math

    # Streaming, so we can iterate over the response.
    r = requests.get(url, stream=True)

    # Total size in bytes.
    total_size = int(r.headers.get('content-length', 0))
    block_size = 1024 * 1024
    wrote = 0
    with open(outfile, 'wb') as f:
        for data in tqdm(r.iter_content(block_size), total=math.ceil(total_size // block_size),
                         unit='MB', unit_scale=True):
            wrote = wrote + len(data)
            f.write(data)
    if total_size != 0 and wrote != total_size:
        raise Exception("ERROR, something went wrong")

# Download database file from github
base_url = 'https://github.com/joheinze/PriNCe/releases/download/'
release_tag = 'v0.5_alpha_release/'
url = base_url + release_tag + db_fname
if not path.isfile(path.join(data_dir, db_fname)):
    print('Downloading for PriNCe database file {0}.'.format(db_fname))
    if debug_level >= 2:
        print(url)
    _download_file(url, path.join(data_dir, db_fname))
else:
    import h5py
    try:
        with h5py.File(path.join(data_dir, db_fname), 'r') as prince_db:
            db_version = (prince_db.attrs['version'])
    except:
        print(f'Database file {db_fname} corrupted. Retrying download.')
        _download_file(url, path.join(data_dir, db_fname))
    finally:
        with h5py.File(path.join(data_dir, db_fname), 'r') as prince_db:
            db_version = (prince_db.attrs['version'])
        if debug_level >= 2:
            print(f'Using database file version {db_version}.')

# if path.isfile(path.join(data_dir, '...previous db name...')):
#     import os
#     print('Removing previous database {0}.'.format('...previous db name...'))
#     os.unlink(path.join(data_dir, '...previous db name...'))




from os.path import join

import numpy as np

from prince_cr.data import spec_data
from prince_cr.util import info, bin_widths, get_AZN
import prince_cr.config as config

from .base import CrossSectionBase
class SophiaSuperposition(CrossSectionBase):
    """ Cross sections generated using the Sophia event generator for protons and neutrons.
    Includes redistribution functions into secondaries

    Cross section for Nuclei are built as a superposition:
    csec_(Z,A) = Z * csec_P + (A - Z) * csec_N

    WARNING: This will not check, if the requested nucleus is existent in nature!
    """

    def __init__(self, *args, **kwargs):
        # Tell the interpolator that this model contains the necessary features
        # for energy redistribution functions
        self.supports_redistributions = True
        CrossSectionBase.__init__(self)
        self._load()

    def _load(self):
        from prince_cr.data import db_handler
        info(2, "Load tabulated cross sections")
        photo_nuclear_tables = db_handler.photo_meson_db('SOPHIA')
        info(2, "Loading SOPHIA cross sections from file.")

        egrid = photo_nuclear_tables["energy_grid"]
        xbins = photo_nuclear_tables["xbins"]
        info(2, "Egrid loading finished")

        # Integer idices of mothers and inclusive channels are stored
        # in first column(s)
        pid_nonel = photo_nuclear_tables["inel_mothers"]
        pids_incl = photo_nuclear_tables["mothers_daughters"]

        # the rest of the line denotes the crosssection on the egrid in mbarn,
        # which is converted here to cm^2
        nonel_raw = photo_nuclear_tables["inelastic_cross_sctions"]
        incl_raw = photo_nuclear_tables["fragment_yields"]

        info(2, "Data file loading finished")

        self._egrid_tab = egrid
        self.cs_proton_grid = nonel_raw[pid_nonel==101].flatten()
        self.cs_neutron_grid = nonel_raw[pid_nonel==100].flatten()

        self.xbins = xbins
        self.redist_proton = {}
        self.redist_neutron = {}
        for (mo, da), csgrid in zip(pids_incl, incl_raw):
            if mo == 101:
                self.redist_proton[da] = csgrid
            elif mo == 100:
                self.redist_neutron[da] = csgrid
            else:
                raise Exception(f'Sophia model should only contain protons and neutrons, but has mother id {mo}')

        # set up inclusive differential channels for protons and neutron
        # The model can return both, integrated over x and redistributed.
        for da in sorted(self.redist_proton):
            self.incl_diff_idcs.append((101, da))
            #self.incl_idcs.append((101, da))
        for da in sorted(self.redist_neutron):
            self.incl_diff_idcs.append((100, da))
            #self.incl_idcs.append((100, da))

        # For more convenient generation of trivial redistribution matrices when joining
        self.redist_shape = (self.xbins.shape[0], self._egrid_tab.shape[0])
        self.set_range()

    def generate_incl_channels(self, mo_indices):
        """Generates indices for all allowed channels given mo_indices
        
        Args:
            mo_indices (list of ints): list of indices for mother nuclei

        Returns:
           Returns:
            list of tuples: list of allowed channels given as (mo_idx, da_idx)
        """
        incl_channels = []
        #return incl_channels

        # This model is a Superposition of protons and neutrons, so we need all respective daughters
        for idx in mo_indices:
            # if idx > 200:
            #     continue
            # add all daughters that are allowed for protons
            for da in sorted(self.redist_proton):
                incl_channels.append((idx, da))
            # add all daughters that are allowed for neutrons
            for da in sorted(self.redist_neutron):
                incl_channels.append((idx, da))
            # add residual nucleus to channel lists
            # NOTE: Including residual nuclei that are not in the disintegration model
            #       Caused some problems, therefore we ignore them here. The effect is
            #       minor though, as photo-meson is subdominant
            # for da in [idx - 101, idx - 100]:
            #    if da > 199:
            #        incl_channels.append((idx, da))

        self.incl_diff_idcs = sorted(list(set(self.incl_diff_idcs + incl_channels)))

        return incl_channels

    def nonel(self, mother):
        r"""Returns non-elastic cross section.

        Absorption cross section of `mother`, which is
        the total minus elastic, or in other words, the inelastic
        cross section.

        Args:
            mother (int): Mother nucleus(on)

        Returns:
           Returns:
            (numpy.array, numpy.array): self._egrid_tab (:math:`\epsilon_r`),
            nonelastic (total) cross section in :math:`cm^{-2}`
        """

        # now interpolate these as Spline
        _, Z, N = get_AZN(mother)

        # the nonelastic crosssection is just a superposition of
        # the proton/neutron number
        cgrid = Z * self.cs_proton_grid + N * self.cs_neutron_grid
        return self.egrid, cgrid[self._range]

    def incl(self, mother, daughter):
        r"""Returns inclusive cross section.

        Inclusive cross section for daughter in photo-nuclear
        interactions of `mother`.

        Args:
            mother (int): Mother nucleus(on)
            daughter (int): Daughter nucleus(on)

        Returns:
            (numpy.array, numpy.array): self._egrid_tab (:math:`\epsilon_r`),
            inclusive cross section in :math:`cm^{-2}`
        """

        _, Z, N = get_AZN(mother)

        if daughter <= 101:
            # raise Exception('Boost conserving cross section called ' +
            #                 'for redistributed particle')
            from scipy.integrate import trapz

            _, cs_diff = self.incl_diff(mother, daughter)
            cs_incl = trapz(cs_diff, x=self.xcenters,
                            dx=bin_widths(self.xbins), axis=0)
            return self.egrid, cs_incl[self._range]

        elif daughter >= 200 and daughter not in [mother - 101, mother - 100]:
            info(10, 'mother, daughter', mother, daughter, 'out of range')
            return self.egrid[[0, -1]], np.array([0., 0.])

        if daughter in [mother - 101]:
            cgrid = Z * self.cs_proton_grid
            # created incl. diff. index for all particle created in p-gamma
            for da in self.redist_proton:
                self.incl_diff_idcs.append((mother, da))
            return self.egrid, cgrid[self._range]
        elif daughter in [mother - 100]:
            cgrid = N * self.cs_neutron_grid
            # created incl. diff. channel index for all particle created in n-gamma
            for da in self.redist_neutron:
                self.incl_diff_idcs.append((mother, da))
            return self.egrid, cgrid[self._range]
        else:
            raise Exception(
                'Channel {:} to {:} not allowed in this superposition model'.
                format(mother, daughter))

    def incl_diff(self, mother, daughter):
        r"""Returns inclusive differential cross section.

        Inclusive differential cross section for daughter in photo-nuclear
        interactions of `mother`. Only defined, if the daughter is distributed
        in :math:`x_{\rm L} = E_{da} / E_{mo}`

        Args:
            mother (int): Mother nucleus(on)
            daughter (int): Daughter nucleus(on)

        Returns:
            (numpy.array, numpy.array, numpy.array): :math:`\epsilon_r` grid,
            :math:`x` grid, differential cross section in :math:`{\rm cm}^{-2}`
        """

        _, Z, N = get_AZN(mother)

        if daughter > 101:
            raise Exception(
                'Redistribution function requested for boost conserving particle'
            )
        csec_diff = None
        # TODO: File shall contain the functions in .T directly
        #   JH: I left it like this on purpose, since the raw data is ordered more consistently
        #       i.e. redist.shape = cs_nonel.shape + xbins.shape
        #       The ordering should rather be changed in the rest of the code
        if daughter in self.redist_proton:
            csec_diff = self.redist_proton[daughter].T * Z

        if daughter in self.redist_neutron:
            cgrid = N * self.cs_neutron_grid
            if np.any(csec_diff):
                csec_diff += self.redist_neutron[daughter].T * N
            else:
                csec_diff = self.redist_neutron[daughter].T * N

        return self.egrid, csec_diff[:, self._range]



class EmpiricalModel(SophiaSuperposition):
    """Photomeson model based on empirical relations and experimental data.
    More info:  L. Morejon et al JCAP11(2019)007
    """

    def __init__(self, *args, **kwargs):
        """Initialize object based on a universal function, and a 
        scaling function.
        """
        SophiaSuperposition.__init__(self)

        self._load_universal_function()

        self._load_pion_function()

        self._fill_multiplicity()

        # self._optimize_and_generate_index() # without this works!!

    def A_eff(self, A, x):
        """Returns the effective A using precomputed scaling coefficients based on
        experimental data.
        """
        from scipy.interpolate import interp1d

        def sigm(x, shift=0., gap=1, speed=1, base=0., rising=False):
            """Models a general sigmoid with multiple parameters.

            Parameters:
            -----------
            x: x values, argument
            shift: middle point, inflection point
            gap: maximal value - minimal value
            speed: controls the speed of the change, 
            base: minimal value
            """
            sigmoid = 1. /(1 + np.exp(- speed * (x - shift)))
            
            if rising:
                return gap * sigmoid + base
            else:
                return gap*( 1. - sigmoid) + base

        mass_scaling_file = join(config.data_dir, 'scaling_lines')
        nsc = np.load(mass_scaling_file, allow_pickle=True,encoding='latin1')

        # defining the transistion smoothing function s 
        # emid = np.array(range(600, 630, 10) + range(680, 720, 10))
        # ymid = np.where(emid < 650, np.interp(emid, nsc[0], nsc[1]), .66)
        # s = interp1d(emid, ymid, 'slinear')

        y = []
        for xi in x:
            if xi < 7:
                y.append(sigm(xi, .7, .099, .7, .942))
            elif xi < 300:
                y.append(np.interp(xi, nsc[0], nsc[1]))
            else:
                y.append(sigm(xi, 430, .2, .01, .66))

        return A**np.array(y)

    def fade(self, cs1, cs2, indices=None):
        """Smoothes the transition from cs1 to cs2
        Uses a sigmoid to smothen the transition between differential 
        cross sections cs1 and cs2 in the energy range defined by the
        indices.
        """
        if indices is None:
            return cs1
            
        x = -100 * np.ones_like(cs1)
        x[..., indices[-1]:] = 100
        x[..., indices] = np.linspace(-5, 5, len(indices))
        
        def sigmoid(x):
            return 1./(1 + np.exp(-x))

        return cs1[..., :]*(1 - sigmoid(x)) + cs2[..., :]*sigmoid(x)

    def _load_universal_function(self):
        """Returns the universal function on a fixed energy range
        """
        from pickle import load as pickle_load
        from scipy.interpolate import UnivariateSpline

        uf_file = join(config.data_dir, 'universal-spline.pkl')
        with open(uf_file, 'rb') as f:
            tck = pickle_load(f,encoding='latin1') 

        self.univ_spl = UnivariateSpline._from_tck(tck)

    def _load_pion_function(self):
        """Returns the universal function on a fixed energy range
        """
        from pickle import load as pickle_load
        from scipy.interpolate import UnivariateSpline

        uf_file = join(config.data_dir, 'pion_spline.pkl')
        with open(uf_file, 'rb') as f:
            tck = pickle_load(f,encoding='latin1') 

        self.pion_spl = UnivariateSpline._from_tck(tck)

    def _fill_multiplicity(self, *args, **kwargs):
        """Populates all tabs (_nonel_tab, _incl_tab, _incl_diff_tab) so they can work
        with the _optimize_indices() method of the base class (CrossSectionBase).
        """
        from ._phenom_relations import multiplicity_table

        self._nonel_tab = {100:(), 101:()}
        
        for mom in self._nonel_tab:
            for dau in [2, 3, 4, 100, 101]:
                self._incl_diff_tab[mom, dau] = ()
                
        new_multiplicity = {}
        nuclides = sorted([k for k in spec_data.keys() if isinstance(k, int)])
        for mom in nuclides:
            A, _, _ = get_AZN(mom)
            if (mom < 101) or (A > config.max_mass) or \
                isinstance(mom, str) or (spec_data[mom]['lifetime'] < config.tau_dec_threshold):
                continue
            mults = multiplicity_table(mom)
            # dau_list, csincl_list = zip(*((k, v) for k, v in mults.iteritems()))
            
            self._nonel_tab[mom] = ()
            for dau in [2, 3, 4, 100, 101]:
                self._incl_diff_tab[mom, dau] = SophiaSuperposition.incl_diff(self, mom, dau)[1]
            
            for dau, mult in mults.items():
                new_multiplicity[mom, dau] = mult
                self._incl_tab[mom, dau] = np.array([])
            
        self.multiplicity = new_multiplicity

    def nonel(self, mother):
        """Computes nonelasatic as A * universal_funtion below 1.9GeV
        and as SophiaSuperposition with given scaling betond 1.9GeV
        """
        e_max = 1.2  # prefixed based on data
        e_scale = .3  # prefixed based on data
        A, _, _ = get_AZN(mother)
        egrid, csnonel = SophiaSuperposition.nonel(self, mother)

        csnonel[egrid <= e_max] = A * self.univ_spl(egrid[egrid <= e_max])
        csnonel[egrid > e_scale] = (csnonel * self.A_eff(A, egrid)/A)[egrid > e_scale]

        return egrid, csnonel

    def incl(self, mother, daughter):
        """Computes inclusive from nonel * M with M is
        multiplicity value stored in internal table
        """
        from scipy.integrate import trapz

        if daughter <= config.redist_threshold_ID:
            _, cs_diff = self.incl_diff(mother, daughter)
            cs_incl = trapz(cs_diff, x=self.xcenters,
                            dx=bin_widths(self.xbins), axis=0)
            return self.egrid, cs_incl[self._range]
        elif (mother, daughter) in self.multiplicity:
            egrid, cs_nonel = self.nonel(mother)
            cs_nonel *= self.multiplicity[mother, daughter]
            return egrid, cs_nonel
        else:
            return SophiaSuperposition.incl(self, mother, daughter)

    def incl_diff(self, mother, daughter):
        """Uses corresponding method from SophiaSuperposition class,
        adding nonel to increase multiplicity by one
        """
        egrid, cs_diff = SophiaSuperposition.incl_diff(self, mother,
                                                       daughter)
        if (mother, daughter) in self.multiplicity:
            xw = self.xwidths[-1]  # accounting for bin width
            _, cs_nonel = self.nonel(mother)
            cs_diff[-1, :] += \
                self.multiplicity[mother, daughter] * cs_nonel / xw
        elif (mother > 101) and (daughter in [2, 3, 4]):  # if it's a pion rescale to A^2/3
            def superposition_incl(mother, daughter):
                from scipy.integrate import trapz
                _, Z, N = get_AZN(mother)
                cs_diff = self.redist_proton[daughter].T * Z * self.cs_proton_grid + \
                    self.redist_neutron[daughter].T * N * self.cs_neutron_grid
                cs_incl = trapz(cs_diff, x=self.xcenters,
                                dx=bin_widths(self.xbins), axis=0)
                return cs_incl[self._range]

            def superposition_multiplicities(mother ,daughter):
                _, Z, N = get_AZN(mother)
                cs_incl = superposition_incl(mother, daughter)
                cs_nonel = Z * self.cs_proton_grid + N * self.cs_neutron_grid
                return cs_incl / cs_nonel[self._range]

            Am, _, _ = get_AZN(mother)
            cs_diff *= float(Am)**(-1/3.)  # ... rescaling SpM to A^2/3

            cs_incl_pi0_sophia = superposition_incl(mother, daughter) * float(Am)**(-1/3.)
            cs_incl_pi0_data = 1e-30*self.pion_spl(egrid * 1e3)*Am**(2./3)

            M_pi = superposition_multiplicities(mother, daughter)
            M_pi0 = superposition_multiplicities(mother, 4)

            renorm = M_pi / M_pi0 * cs_incl_pi0_data / cs_incl_pi0_sophia
            cs_diff_renormed = cs_diff * renorm
            cs_diff_renormed = self.fade(cs_diff, cs_diff_renormed, range(32)) # hardcoded index, found manually
            cs_diff = self.fade(cs_diff_renormed, cs_diff, range(55, 95)) # hardcoded index, found manually

            # # additional correction to pion scaling high energies, after paper was corrected
            # def sigm(x, shift=0., gap=1, speed=1, base=0., rising=False):
            #     """Models a general sigmoid with multiple parameters.

            #     Parameters:
            #     -----------
            #     x: x values, argument
            #     shift: middle point, inflection point
            #     gap: maximal value - minimal value
            #     speed: controls the speed of the change, 
            #     base: minimal value
            #     """
            #     sigmoid = 1. /(1 + np.exp(- speed * (x - shift)))
                
            #     if rising:
            #         return gap * sigmoid + base
            #     else:
            #         return gap*( 1. - sigmoid) + base

            alpha_plus = np.where(egrid <= 1., 2./3, 1 - np.exp( -4./7*(egrid - 1)**.5)/3)
            cs_diff *= Am**(alpha_plus - 2./3)

        return egrid, cs_diff



"""The module contains everything to handle cross section interfaces."""

from abc import ABCMeta, abstractmethod
from os.path import isfile, join

import numpy as np

import prince_cr.decays as decs
from prince_cr.data import spec_data
from prince_cr.util import (
    get_2Dinterp_object, get_interp_object, info, bin_widths, get_AZN)
import prince_cr.config as config

class CrossSectionBase(object, metaclass=ABCMeta):
    """Base class for cross section interfaces to tabulated models.

    The class is abstract and it is not inteded to be instantiated.

    Child Classes either define the tables:
        self._egrid_tab
        self._nonel_tab
        self._incl_tab
        self._incl_diff

    Or directly reimplememnt the functions
        nonel(self, mother, daughter)
        incl(self, mother, daughter)
        incl_diff(self, mother, daughter)

    The flag self.supports_redistributions = True/False should be set
    To tell the class to include/ignore incl_diff
    """
    def __init__(self):
        # Tuple, defining min and max energy cuts on the grid
        self._range = None
        # Energy grid, as defined in files
        self._egrid_tab = None
        # Dictionary of nonel. cross sections on egrid, indexed by (mother)
        self._nonel_tab = {}
        # Dictionary of incl. cross sections on egrid, indexed by (mother, daughter)
        self._incl_tab = {}
        # Dictionary of incl. diff. cross sections on egrid, indexed by (mother, daughter)
        self._incl_diff_tab = {}
        # List of available mothers for nonel cross sections
        self.nonel_idcs = []
        # List of available (mothers,daughter) reactions in incl. cross sections
        self.incl_idcs = []
        # List of available (mothers,daughter) reactions in incl. diff. cross sections
        self.incl_diff_idcs = []
        # Common grid in x (the redistribution variable)
        self.xbins = None

        # Flag, which tells if the model supports secondary redistributions
        if not hasattr(self, 'supports_redistributions'):
            self.supports_redistributions = None  # JH: to differ from explicitly set False
        # List of all known particles (after optimization)
        self.known_species = []
        # List of all boost conserving inclusive channels (after optimization)
        self.known_bc_channels = []
        # List of all differential inclusive channels (after optimization)
        self.known_diff_channels = []
        # Dictionary of (mother, daughter) reactions for each mother
        self.reactions = {}

        # Class name of the model
        self.mname = self.__class__.__name__

    def set_range(self, e_min=None, e_max=None):
        """Set energy range within which to return tabulated data.

        Args:
            e_min (float): minimal energy in GeV
            e_max (float): maximal energy in GeV
        """
        if e_min is None:
            e_min = np.min(self._egrid_tab)
        if e_max is None:
            e_max = np.max(self._egrid_tab)

        info(5, "Setting range to {0:3.2e} - {1:3.2e}".format(e_min, e_max))
        self._range = np.where((self._egrid_tab >= e_min)
                               & (self._egrid_tab <= e_max))[0]
        info(
            2, "Range set to {0:3.2e} - {1:3.2e}".format(
                np.min(self._egrid_tab[self._range]),
                np.max(self._egrid_tab[self._range])))

    @property
    def egrid(self):
        """Returns energy grid of the tabulated data in selected range.

        Returns:
            (numpy.array): Energy grid in GeV
        """

        return self._egrid_tab[self._range]

    @property
    def xcenters(self):
        """Returns centers of the grid in x.

        Returns:
            (numpy.array): x grid
        """

        return 0.5 * (self.xbins[1:] + self.xbins[:-1])

    @property
    def xwidths(self):
        """Returns bin widths of the grid in x.

        Returns:
            (numpy.array): x widths
        """

        return self.xbins[1:] - self.xbins[:-1]

    @property
    def resp(self):
        """Return ResponseFunction corresponding to this cross section
        Will only create the Response function once. 
        """
        if not hasattr(self, '_resp'):
            info(2, 'First Call, creating instance of ResponseFunction now')
            from .response import ResponseFunction
            self._resp = ResponseFunction(self)
        return self._resp

    def is_differential(self, mother, daughter):
        """Returns true if the model supports redistributions and requested
        mother/daughter combination should return non-zero redistribution matrices.

        Args:
            mother (bool): Neucosma ID of mother particle
            daughter (bool): Neucosma ID of daughter particle

        Returns:
            (bool): ``True`` if the model has this particular redistribution function
        """
        # info(10, mother, daughter, " asking for redist")
        # if not self.supports_redistributions:
        #     info(10, mother, daughter, " model doesn't support redist")
        #     return False
        if (daughter <= config.redist_threshold_ID
                or (mother, daughter) in self.incl_diff_idcs):
            info(60, 'Daughter requires redistribution.', mother, daughter)
            return True
        info(60, 'Daughter conserves boost.', mother, daughter)
        return False

    def _update_indices(self):
        """Updates the list of indices according to entries in the
        _tab variables"""

        self.nonel_idcs = sorted(self._nonel_tab.keys())
        self.incl_idcs = sorted(self._incl_tab.keys())
        self.incl_diff_idcs = sorted(self._incl_diff_tab.keys())

    def generate_incl_channels(self, mo_indices):
        """Generates indices for all allowed channels given mo_indices
            Note: By default this returns an empty list,
                  meant to be overwritten in cases where 
                  the child class needs to dynamically generate indices
        
        Args:
            mo_indices (list of ints): list of indices for mother nuclei

        Returns:
           Returns:
            list of tuples: list of allowed channels given as (mo_idx, da_idx)
        """
        incl_channels = []

        return incl_channels

    def _optimize_and_generate_index(self):
        """Construct a list of mothers and (mother, daughter) indices.

        Args:
            just_reactions (bool): If True then fill just the reactions index.
        """

        # Integrate out short lived processes and leave only stable particles
        # in the databases
        self._reduce_channels()

        # Go through all three cross section categories
        # index contents in the ..known..variable
        self.reactions = {}

        self._update_indices()

        for mo, da in self.incl_idcs:
            if da >= 100 and get_AZN(da)[0] > get_AZN(mo)[0]:
                raise Exception(
                    'Daughter {0} heavier than mother {1}. Physics??'.format(
                        da, mo))

            if mo not in self.reactions:
                self.reactions[mo] = []
                self.known_species.append(mo)

            if (mo, da) not in self.reactions[mo]:
                # Make sure it's a unique list
                self.reactions[mo].append((mo, da))
            if self.is_differential(mo, da):
                # Move the distributions which are expected to be differential
                # to _incl_diff_tab
                self._incl_diff_tab[(mo, da)] = self._arange_on_xgrid(
                    self._incl_tab.pop((mo, da)))
                info(10, "Channel {0} -> {1} forced to be differential.")
            else:
                self.known_bc_channels.append((mo, da))
                self.known_species.append(da)

        for mo, da in list(self._incl_diff_tab.keys()):
            if da >= 100 and get_AZN(da)[0] > get_AZN(mo)[0]:
                raise Exception(
                    'Daughter {0} heavier than mother {1}. Physics??'.format(
                        da, mo))

            if mo not in self.reactions:
                self.reactions[mo] = []
                self.known_species.append(mo)

            if (mo, da) not in self.reactions[mo]:
                # Make sure it's a unique list to avoid unnecessary loops
                self.reactions[mo].append((mo, da))
                self.known_diff_channels.append((mo, da))
                self.known_species.append(da)

        # Remove duplicates
        self.known_species = sorted(list(set(self.known_species)))
        self.known_bc_channels = sorted(list(set(self.known_bc_channels)))
        self.known_diff_channels = sorted(list(set(self.known_diff_channels)))

        for sp in self.known_species:
            if sp >= 100 and (sp, sp) not in self.known_diff_channels:
                self.known_bc_channels.append((mo, mo))
            if (mo, mo) not in self.reactions[mo]:
                self.reactions[mo].append((mo, mo))

        # Make sure the indices are up to date
        self._update_indices()

        # Count numbers of channels for statistics
        # Count number of incl channels for activated nuclear species
        # n_incl = np.sum([
        #     len(self.reactions[mother])
        #     for mother in self.spec_man.known_species if mother >= 100
        # ])

    def _reduce_channels(self):
        """Follows decay chains until all inclusive reactions point to
        stable final state particles.

        The "tau_dec_threshold" parameter in the config controls the
        definition of stable. Unstable nuclei for which no decay channels
        are known, will be forced to beta-decay until they reach a stable
        element.
        """
        from prince_cr.util import AdditiveDictionary
        # TODO: check routine, how to avoid empty channels and
        # mothers with zero nonel cross sections

        # The new dictionary that will replace _incl_tab
        new_incl_tab = AdditiveDictionary()
        new_dec_diff_tab = AdditiveDictionary()

        threshold = config.tau_dec_threshold

        # How to indent debug printout for recursion
        dbg_indent = lambda lev: 4 * lev * "-" + ">" if lev else ""

        info(2, "Integrating out species with lifetime smaller than",
             threshold)
        info(3, (
            "Before optimization, the number of known primaries is {0} with " +
            "in total {1} inclusive channels").format(len(self._nonel_tab),
                                                      len(self._incl_tab)))

        if self.xbins is None:
            info(
                4,
                'Model does not provide a native xbins. Assuming JH special sophia',
                'binning.')
            from .photo_meson import SophiaSuperposition
            self.xbins = SophiaSuperposition().xbins

        bc = self.xcenters
        bw = bin_widths(self.xbins)
        # The x_mu/x_pi grid
        # dec_grid = np.fromfunction(
        #     lambda j, i: 10**(np.log10(bc[1] / bc[0]) * (j - i)), (len(bc),
        #                                                            len(bc)))

        # dec_grid = np.outer(bc, 1 / bc)

        dec_bins = np.outer(self.xbins, 1 / bc)
        dec_bins_lower = dec_bins[:-1]
        dec_bins_upper = dec_bins[1:]

        # dec_grid[dec_grid > 1.] *= 0.
        # The differential element dx_mu/x_pi
        int_scale = np.tile(bw / bc, (len(bc), 1))

        from functools import lru_cache
        @lru_cache(maxsize=512, typed=False)
        def decay_cached(mother,daughter):
            dec_dist = int_scale * decs.get_decay_matrix_bin_average(
                mother, daughter, dec_bins_lower, dec_bins_upper)
            
            return dec_dist

        def convolve_with_decay_distribution(diff_dist, mother, daughter,
                                             branching_ratio):
            r"""Computes the prompt decay xdist by convolving the x distribution
            of the unstable particle with the decay product distribution.

            :math:`\frac{{\rm d}N^{A\gamma \to \mu}}{{\rm d}x_j} = 
            \sum_{i=0}^{N_x}~\Delta x_i 
            \frac{{\rm d}N^{A\gamma \to \pi}}{{\rm d} x_i}~
            \frac{{\rm d}N^{\pi \to \mu}}{{\rm d} x_j}`
            """
            # dec_dist = int_scale * decs.get_decay_matrix(
            #     mother, daughter, dec_grid)
            dec_dist = decay_cached(mother,daughter)

            info(20, 'convolving with decay dist', mother, daughter)
            # Handle the case where table entry is (energy_grid, matrix)
            if not isinstance(diff_dist, tuple):
                return branching_ratio * dec_dist.dot(diff_dist)
            else:
                return diff_dist[0], branching_ratio * dec_dist.dot(
                    diff_dist[1])

        def follow_chain(first_mo, da, csection, reclev):
            """Recursive function to follow decay chains until all
            final state particles are stable.
            
            The result is saved in two dictionaries; one for the boost
            conserving inclusive channels and the other one collects
            channels with meson or lepton decay products, which will
            need special care due to energy redistributions of these
            secondaries.
            """

            info(10, dbg_indent(reclev), 'Entering with', first_mo, da)

            if da not in spec_data:
                info(
                    3, dbg_indent(reclev),
                    'daughter {0} unknown, forcing beta decay. Not Implemented yet!!'
                    .format(da))
                return

            # Daughter is stable. Add it to the new dictionary and terminate
            # recursion
            if spec_data[da]["lifetime"] >= threshold:
                if self.is_differential(None, da):
                    # If the daughter is a meson or lepton, use the dictionary for
                    # differential channels
                    info(
                        20, dbg_indent(reclev),
                        'daughter {0} stable and differential. Adding to ({1}, {2})'
                        .format(da, first_mo, da))
                    new_dec_diff_tab[(first_mo, da)] = csection
                else:
                    info(
                        20, dbg_indent(reclev),
                        'daughter {0} stable. Adding to ({1}, {2})'.format(
                            da, first_mo, da))
                    new_incl_tab[(first_mo, da)] = csection
                return

            # ..otherwise follow decay products of this daughter, tracking the
            # original mother particle (first_mo). The cross section (csection) is
            # reduced by the branching ratio (br) of this particular channel
            for br, daughters in spec_data[da]["branchings"]:
                info(10, dbg_indent(reclev),
                     ("{3} -> {0:4d} -> {2:4.2f}: {1}").format(
                         da, ", ".join(map(str, daughters)), br, first_mo))

                for chained_daughter in daughters:
                    # Follow each secondary and increment the recursion level by one
                    if self.is_differential(None, chained_daughter):
                        info(10, 'daughter', chained_daughter, 'of', da,
                             'is differential')
                        follow_chain(
                            first_mo, chained_daughter,
                            convolve_with_decay_distribution(
                                self._arange_on_xgrid(csection), da,
                                chained_daughter, br), reclev + 1)
                    else:
                        follow_chain(first_mo, chained_daughter, br * csection,
                                     reclev + 1)

        # Remove all unstable particles from the dictionaries
        for mother in sorted(self._nonel_tab.keys()):
            if mother not in spec_data or spec_data[mother][
                    "lifetime"] < threshold:
                info(
                    20,
                    "Primary species {0} does not fulfill stability criteria.".
                    format(mother))
                _ = self._nonel_tab.pop(mother)
        # Only stable (interacting) mother particles are left
        self._update_indices()

        for (mother, daughter) in self.incl_idcs:

            if mother not in self.nonel_idcs:
                info(
                    30, "Removing {0}/{1} from incl, since mother not stable ".
                    format(mother, daughter))
                _ = self._incl_tab.pop((mother, daughter))

            elif self.is_differential(mother, daughter):
                # Move the distributions which are expected to be differential
                # to _incl_diff_tab
                self._incl_diff_tab[(mother,
                                     daughter)] = self._arange_on_xgrid(
                                         self._incl_tab.pop(
                                             (mother, daughter)))

        self._update_indices()

        for (mother, daughter) in self.incl_diff_idcs:

            if mother not in self.nonel_idcs:
                info(
                    30,
                    "Removing {0}/{1} from diff incl, since mother not stable "
                    .format(mother, daughter))
                _ = self._incl_diff_tab.pop((mother, daughter))

        self._update_indices()

        # Launch the reduction for each inclusive channel
        for (mo, da), value in list(self._incl_tab.items()):
            #print mo, da, value
            #print '---'*30
            follow_chain(mo, da, value, 0)

        for (mo, da), value in list(self._incl_diff_tab.items()):
            #print mo, da, value
            #print '---'*30
            follow_chain(mo, da, value, 0)

        # Overwrite the old incl dictionary
        self._incl_tab = dict(new_incl_tab)
        # Overwrite the old incl_diff dictionary
        self._incl_diff_tab = dict(new_dec_diff_tab)
        # Reduce also the incl_diff_tab by removing the unknown mothers. At this stage
        # of the code, the particles with redistributions are
        info(
            3,
            ("After optimization, the number of known primaries is {0} with " +
             "in total {1} inclusive channels").format(
                 len(self._nonel_tab),
                 len(self._incl_tab) + len(self._incl_diff_tab)))
        info(2, f'Cache used for decays, {decay_cached.cache_info()}') # pylint:disable=no-value-for-parameter

    def nonel_scale(self, mother, scale='A'):
        """Returns the nonel cross section scaled by `scale`.

        Convenience funtion for plotting, where it is important to
        compare the cross section per nucleon.

        Args:
            mother (int): Mother nucleus(on)
            scale (float): If `A` then nonel/A is returned, otherwise
                           scale can be any float.

        Returns:
            (numpy.array, numpy.array): Tuple of Energy grid in GeV,
                                        scale * inclusive cross section
                                        in :math:`cm^{-2}`
        """

        egr, csection = self.nonel(mother)

        if scale == 'A':
            scale = 1. / get_AZN(mother)[0]

        return egr, scale * csection

    def incl_scale(self, mother, daughter, scale='A'):
        """Same as :func:`~cross_sections.CrossSectionBase.nonel_scale`,
        just for inclusive cross sections.
        """

        egr, csection = self.incl(mother, daughter)

        if scale == 'A':
            scale = 1. / get_AZN(mother)[0]

        return egr, scale * csection

    def nonel(self, mother):
        """Returns non-elastic cross section.

        Absorption cross section of `mother`, which is
        the total minus elastic, or in other words, the inelastic
        cross section.

        Args:
            mother (int): Mother nucleus(on)

        Returns:
            (numpy.array, numpy.array): Tuple of Energy grid in GeV, inclusive cross
                                        section in :math:`cm^{-2}`
        """

        if mother not in self._nonel_tab:
            raise Exception('Mother {0} unknown.'.format(mother))

        if isinstance(self._nonel_tab[mother], tuple):
            return self._nonel_tab[mother]
        else:
            return self.egrid, self._nonel_tab[mother][self._range]

    def incl(self, mother, daughter):
        """Returns inclusive cross section.

        Inclusive cross section for daughter in photo-nuclear
        interactions of `mother`.

        Args:
            mother (int): Mother nucleus(on)
            daughter (int): Daughter nucleus(on)

        Returns:
            (numpy.array): Inclusive cross section in :math:`cm^{-2}`
                           on self._egrid_tab
        """

        from scipy.integrate import trapz

        if (mother, daughter) in self._incl_diff_tab:
            # Return the integral of the differential for the inclusive
            egr_incl, cs_diff = self.incl_diff(mother, daughter)
            # diff_mat = diff_mat.transpose()
            cs_incl = trapz(cs_diff,
                            x=self.xcenters,
                            dx=bin_widths(self.xbins),
                            axis=0)

            if isinstance(self._incl_diff_tab[(mother, daughter)], tuple):
                return egr_incl, cs_incl

            return self.egrid, cs_incl[self._range]

        elif (mother, daughter) not in self._incl_tab:
            raise Exception(
                self.__class__.__name__ + '::'
                '({0},{1}) combination not in inclusive cross sections'.format(
                    mother, daughter))

        # If _nonel_tab contains tuples of (egrid, cs) return tuple
        # otherwise return (egrid, cs) in range defined by self.range

        if isinstance(self._incl_tab[(mother, daughter)], tuple):
            return self._incl_tab[(mother, daughter)]
        return self.egrid, self._incl_tab[(mother, daughter)][self._range]

    def incl_diff(self, mother, daughter):
        """Returns inclusive cross section.

        Inclusive differential cross section for daughter in photo-nuclear
        interactions of `mother`. Only defined, if the daughter is distributed 
        in :math:`x = E_{da} / E_{mo}`

        Args:
            mother (int): Mother nucleus(on)
            daughter (int): Daughter nucleus(on)

        Returns:
            (numpy.array): Inclusive cross section in :math:`cm^{-2}`
                           on self._egrid_tab
        """

        if (mother, daughter) not in self._incl_diff_tab:
            raise Exception(
                self.__class__.__name__ +
                '({0},{1}) combination not in inclusive differential cross sections'
                .format(mother, daughter))

        # If _nonel_tab contains tuples of (egrid, cs) return tuple
        # otherwise return (egrid, cs) in range defined by self.range

        if isinstance(self._incl_diff_tab[(mother, daughter)], tuple):
            return self._incl_diff_tab[(mother, daughter)]
        return self.egrid, self._incl_diff_tab[(mother,
                                                daughter)][:, self._range]

    def _arange_on_xgrid(self, incl_cs):
        """Returns the inclusive cross section on an xgrid at x=1."""

        egr, cs = None, None

        if isinstance(incl_cs, tuple):
            egr, cs = incl_cs
        else:
            cs = incl_cs

        nxbins = len(self.xbins) - 1
        if len(cs.shape) > 1 and cs.shape[0] != nxbins:
            raise Exception(
                'One dimensional cross section expected, instead got',
                cs.shape, '\n', cs)
        elif len(cs.shape) == 2 and cs.shape[0] == nxbins:
            info(20, 'Supplied 2D distribution seems to be distributed in x.')
            if isinstance(incl_cs, tuple):
                return egr, cs
            return cs

        csec = np.zeros((nxbins, cs.shape[0]))
        # NOTE: The factor 2 in the following line is a workarround to account for the latter linear interpolation
        #       This is needed because linear spline integral will result in a trapz,
        #       which has half the area of the actual first bin
        corr_factor = 2 * self.xwidths[-1] / (self.xcenters[-1] -
                                              self.xcenters[-2])
        csec[-1, :] = cs / self.xwidths[-1] * corr_factor
        info(
            4,
            'Warning! Workaround to account for linear interpolation in x, factor 2 added!'
        )
        if isinstance(incl_cs, tuple):
            return egr, csec
        return csec

    def multiplicities(self, mother, daughter):
        '''Return the multiplicities from either the inclusive channels, or the
        differential ones integrated by x, as a function of Energy.
        '''
        egrid_incl, cs_incl = self.incl(mother, daughter)
        egrid_nonel, cs_nonel = self.nonel(mother)

        if egrid_incl.shape != egrid_nonel.shape:
            raise Exception('Problem with different grid shapes')

        multiplicities = cs_incl / np.where(cs_nonel == 0, np.inf, cs_nonel)

        return egrid_nonel, multiplicities


if __name__ == "__main__":
    pass



"""The module contains classes for computations of interaction rates"""

import numpy as np
from scipy.integrate import trapz

from prince_cr.data import PRINCE_UNITS
from prince_cr.util import info
import prince_cr.config as config

using_cupy = False
# Use GPU support
if config.has_cupy and config.linear_algebra_backend.lower() == 'cupy':
    import cupy
    from prince_cr.config import mempool
    using_cupy = True

class PhotoNuclearInteractionRate(object):
    """Implementation of photo-hadronic/nuclear interaction rates.
    This Version directly writes the data into a CSC-matrix and only updates the data each time.
    """

    def __init__(self, prince_run=None, with_dense_jac=True, *args, **kwargs):
        info(3, 'creating instance')
        self.with_dense_jac = with_dense_jac

        #: Reference to PhotonField object
        self.photon_field = prince_run.photon_field

        #: Reference to CrossSection object
        self.cross_sections = prince_run.cross_sections

        #: Reference to species manager
        self.spec_man = prince_run.spec_man

        # Initialize grids
        self.e_photon = prince_run.ph_grid
        self.e_cosmicray = prince_run.cr_grid

        # Initialize cache of redshift value
        self._ph_vec_zcache = None
        self._ratemat_zcache = None

        # Initialize the matrices for batch computation
        self._batch_rows = None
        self._batch_cols = None
        self._batch_matrix = None
        self._batch_vec = None
        self.coupling_mat = None
        self.dense_coupling_mat = None

        self._estimate_batch_matrix()
        self._init_matrices()
        self._init_coupling_mat()

    def photon_vector(self, z):
        """Returns photon vector at redshift `z` on photon grid.

        This vector is in fact a matrix of vectors of the interpolated
        photon field with dimensions (dim_cr, xi_steps).

        Args:
            z (float): redshift

        Return value from cache if redshift value didn't change since last call.
        """
        # photon_vector = np.zeros_like(self.e_photon.grid)
        photon_vector = self.photon_field.get_photon_density(
            self.e_photon.grid, z)

        return photon_vector

    def _estimate_batch_matrix(self):
        ''' estimate dimension of the batch matrix'''
        dcr = self.e_cosmicray.d
        dph = self.e_photon.d

        batch_dim = 0
        for specid in self.spec_man.known_species:
            if specid < 100:
                continue
            # Add the main diagonal self-couplings (absoption)
            batch_dim += dcr
            for rtup in self.cross_sections.reactions[specid]:
                # Off main diagonal couplings (reinjection)
                if rtup in self.cross_sections.known_bc_channels:
                    batch_dim += dcr
                elif rtup in self.cross_sections.known_diff_channels:
                    # Only half of the elements can be non-zero (energy conservation)
                    batch_dim += int(dcr**2 / 2) + 1

        info(2, 'Batch matrix dimensions are {0}x{1}'.format(batch_dim, dph))
        self._batch_matrix = np.zeros((batch_dim, dph))
        self._batch_rows = []
        self._batch_cols = []
        info(
            3,
            'Memory usage: {0} MB'.format(self._batch_matrix.nbytes / 1024**2))

    def _init_matrices(self):
        """ A new take on filling the matrices"""

        # Define some short-cuts
        known_species = self.spec_man.known_species[::-1]
        sp_id_ref = self.spec_man.ncoid2sref
        resp = self.cross_sections.resp
        m_pr = PRINCE_UNITS.m_proton

        # Energy variables
        dcr = self.e_cosmicray.d
        dph = self.e_photon.d
        ecr = self.e_cosmicray.grid
        bcr = self.e_cosmicray.bins
        eph = self.e_photon.grid
        bph = self.e_photon.bins
        delta_ec = self.e_cosmicray.widths
        delta_ph = self.e_photon.widths

        # Edges of each CR energy bin and photon energy bin
        elims = np.vstack([bcr[:-1], bcr[1:]])
        plims = np.vstack([bph[:-1], bph[1:]])

        # CR and photon grid indices
        emo_idcs = np.arange(dcr)
        eda_idcs = np.arange(dcr)
        p_idcs = np.arange(dph)

        # values for x and y to cut on:
        x_cut = config.x_cut
        y_cut = config.y_cut
        x_cut_proton = config.x_cut_proton

        ibatch = 0
        import itertools
        spec_iter = itertools.product(known_species, known_species)
        for moid, daid in spec_iter:

            if moid < 100:
                continue
            else:
                info(10, f'Filling channel {moid} -> {daid}')

            has_nonel = moid == daid
            if has_nonel:
                intp_nonel = resp.nonel_intp[moid].antiderivative()

            if (((moid, daid) in self.cross_sections.known_bc_channels) or
                (has_nonel and
                 (moid, daid) not in self.cross_sections.known_diff_channels)):

                has_incl = (moid, daid) in resp.incl_intp
                if has_incl:
                    intp_bc = resp.incl_intp[(moid, daid)].antiderivative()
                else:
                    info(1, 'Inclusive interpolator not found for',
                         (moid, daid))

                if not (has_nonel or has_incl):
                    raise Exception('Channel without interactions:',
                                    (moid, daid))

                # The cross sections need to be evaluated
                # on x = E_{CR,da} / E_{CR,mo} and y = E_ph * E_{CR,mo} / m_proton
                # To vectorize the evaluation, we create outer products using numpy
                # broadcasting:

                emo = ecr
                xl = elims[0] / emo
                xu = elims[1] / emo
                delta_x = delta_ec / emo

                yl = plims[0, None, :] * emo[:, None] / m_pr
                yu = plims[1, None, :] * emo[:, None] / m_pr
                delta_y = delta_ph[None, :] * emo[:, None] / m_pr

                int_fac = (delta_ec[:, None] * delta_ph[None, :] / emo[:, None])
                diff_fac = 1. / delta_x[:, None] / delta_y

                # This takes the average by evaluating the integral and dividing by bin
                # width
                if has_incl:
                    self._batch_matrix[ibatch:ibatch + len(emo), :] = (
                        intp_bc(yu) - intp_bc(yl)) * int_fac * diff_fac
                if has_nonel:
                    self._batch_matrix[ibatch:ibatch + len(emo), :] -= (intp_nonel(
                        yu) - intp_nonel(yl)) * int_fac * diff_fac

                # finally map this to the coupling matrix
                ibatch += len(emo)
                self._batch_rows.append(sp_id_ref[daid].lidx() + eda_idcs)
                self._batch_cols.append(sp_id_ref[moid].lidx() + emo_idcs)

            elif (moid, daid) in self.cross_sections.known_diff_channels:

                has_redist = (moid, daid) in resp.incl_diff_intp
                if has_redist:
                    intp_diff = resp.incl_diff_intp[(moid, daid)]
                    intp_diff_integral = resp.incl_diff_intp_integral[(moid, daid)]
                    intp_nonel = resp.nonel_intp[moid]
                    intp_nonel_antid = resp.nonel_intp[moid].antiderivative()

                    ymin = np.min(intp_diff.get_knots()[1])
                else:
                    raise Exception('This should not occur.')

                ibatch_bf = ibatch
                # generate outer products using broadcasting
                emo = ecr[:, None, None]
                eda = ecr[None, :, None]
                epho = eph[None, None, :]
                target_shape = np.ones_like(emo * eda * epho)

                xl = elims[0, None, :, None] / emo * target_shape
                xu = elims[1, None, :, None] / emo * target_shape
                delta_x = delta_ec[None, :, None] / emo

                yl = plims[0, None, None, :] * emo / m_pr * target_shape
                yu = plims[1, None, None, :] * emo / m_pr * target_shape
                delta_y = delta_ph[None, None, :] * emo / m_pr

                int_fac = (delta_ec[:, None, None] *
                           delta_ph[None, None, :] / emo) * target_shape
                diff_fac = 1. / delta_x / delta_y

                # Generate boolean arrays to cut on xvalues
                if daid == 101:
                    cuts = np.logical_and(xl >= x_cut_proton, xl <= 1)
                else:
                    # or (yu < ymin) or (yl > y_cut)
                    cuts = np.logical_and(xl >= x_cut, xl <= 1)
                cuts = cuts[:, :, 0]

                # # NOTE JH: This is an old version, which brute force vectorizes the integral with numpy
                # # I am leaving this in the comments, in case we want to go back for testing-
                # integrator = np.vectorize(intp_diff.integral)
                # res = integrator(xl[cuts], xu[cuts], yl[cuts], yu[cuts]) * diff_fac[cuts] * int_fac[cuts]

                # This takes the average by evaluating the integral and dividing by bin width
                # intp_diff_integral contains the antiderivate, to to get the integral (xl,yl,xu,yu)
                # we need to substract INT = (0,0,xu,yu) - (0,0,xl,yu) - (0,0,xu,yl) +
                # (0,0,xl,yl)
                res = intp_diff_integral.ev(xu[cuts], yu[cuts])
                res -= intp_diff_integral.ev(xl[cuts], yu[cuts])
                res -= intp_diff_integral.ev(xu[cuts], yl[cuts])
                res += intp_diff_integral.ev(xl[cuts], yl[cuts])
                res *= diff_fac[cuts] * int_fac[cuts]
                res[res < 0] = 0.

                # Since we made cuts on x, we need to make the same cut on the index
                # mapping
                emoidx, edaidx, _ = np.meshgrid(sp_id_ref[moid].lidx() + emo_idcs,
                                                sp_id_ref[daid].lidx() + eda_idcs,
                                                p_idcs, indexing='ij')
                emoidx, edaidx = emoidx[cuts], edaidx[cuts]

                # Now add the nonel interactions on the main diagonal
                if has_nonel:
                    res[emoidx == edaidx] -= (
                        intp_nonel_antid(yu[cuts][emoidx == edaidx]) -
                        intp_nonel_antid(yl[cuts][emoidx == edaidx])) * (
                            diff_fac[cuts][emoidx == edaidx] *
                            int_fac[cuts][emoidx == edaidx])

                # Finally write this to the batch matrix
                self._batch_matrix[ibatch:ibatch + len(emoidx), :] = res
                self._batch_rows.append(edaidx[:, 0])
                self._batch_cols.append(emoidx[:, 0])
                ibatch += len(emoidx)

            else:
                info(20, 'Species combination not included in model', moid,
                     daid)

        self._batch_matrix = self._batch_matrix[:ibatch, :]
        self._batch_rows = np.concatenate(self._batch_rows, axis=None)
        self._batch_cols = np.concatenate(self._batch_cols, axis=None)
        self._batch_vec = np.zeros(ibatch)

        info(2, f'Batch matrix shape: {self._batch_matrix.shape}')
        info(2, f'Batch rows shape: {self._batch_rows.shape}')
        info(2, f'Batch cols shape: {self._batch_cols.shape}')
        info(2, f'Batch vector shape: {self._batch_vec.shape}')

        memory = (self._batch_matrix.nbytes + self._batch_rows.nbytes +
                  self._batch_cols.nbytes + self._batch_vec.nbytes) / 1024**2
        info(3, "Memory usage after initialization: {:} MB".format(memory))

    def _init_coupling_mat(self):
        """Initialises the coupling matrix directly in sparse (csr) format.
        """
        info(0, 'Initiating coupling matrix in ({:}) format'.format('CSR'))

        from scipy.sparse import csr_matrix
        if using_cupy:
            # For GPU we initialize the csr matrix on the host and then cast to GPU
            from cupyx.scipy.sparse import csr_matrix as cp_csr_matrix
            self.coupling_mat_np = csr_matrix(
                (self._batch_vec.astype(np.float32),
                    (self._batch_rows, self._batch_cols)),
                copy=True)
            self.coupling_mat = cp_csr_matrix(self.coupling_mat_np, copy=True)
            self._batch_vec = self.coupling_mat.data
            del self.coupling_mat_np
        else:
            self.coupling_mat = csr_matrix(
                (self._batch_vec, (self._batch_rows, self._batch_cols)),
                copy=True)

        # create an index to sort by rows and then columns,
        # which is the same ordering CSR has internally
        # lexsort sorts by last argument first!!!
        self.sortidx = np.lexsort((self._batch_cols, self._batch_rows))

        self._batch_rows = self._batch_rows[self.sortidx]
        self._batch_cols = self._batch_cols[self.sortidx]

        # Reorder batch matrix according to order in coupling_mat
        if using_cupy:
            self._batch_matrix = cupy.array(
                self._batch_matrix[self.sortidx, :], dtype=np.float32)
        else:
            self._batch_matrix = self._batch_matrix[self.sortidx, :]

    def _update_rates(self, z, force_update=False):
        """Batch compute all nonel and inclusive rates if z changes.

        The result is always stored in the same vectors, since '_init_rate_matstruc'
        makes use of views to link ranges of the vector to locations in the matrix.

        Args:
            z (float): Redshift value at which the photon field is taken.

        Returns:
            (bool): True if fields we indeed updated, False if nothing happened.
        """
        if self._ratemat_zcache != z or force_update:
            info(5, 'Updating batch rate vectors.')

            if using_cupy:
                if isinstance(self._batch_matrix, np.ndarray):
                    self._init_coupling_mat()
                cupy.dot(
                    self._batch_matrix, cupy.array(
                        self.photon_vector(z), dtype=np.float32),
                    out=self._batch_vec)
            else:
                np.dot(
                    self._batch_matrix, self.photon_vector(z), out=self._batch_vec)
            self._ratemat_zcache = z
            return True
        else:
            return False

    def _update_coupling_mat(self, z, scale_fac, force_update=False):
        """Updates the sparse (csr) coupling matrix
        Only the data vector is updated to minimize computation
        """
        # Do not execute dot product if photon field didn't change
        if self._update_rates(z, force_update):
            self.coupling_mat.data = scale_fac * self._batch_vec

    def get_hadr_jacobian(self, z, scale_fac=1., force_update=False):
        """Returns the nonel rate vector and coupling matrix.
        """
        self._update_coupling_mat(z, scale_fac, force_update)
        return self.coupling_mat

    def single_interaction_length(self, pid, z, pfield=None):
        """Returns energy loss length in cm
        (convenience function for plotting)
        """
        if pfield is not None:
            mem_pfield = self.photon_field
            self.photon_field = pfield

        species = self.spec_man.ncoid2sref[pid]
        egrid = self.e_cosmicray.grid * species.A
        rate = -1 * self.get_hadr_jacobian(
            force_update=True, z=z).toarray()[
                species.sl, species.sl].diagonal()

        length = 1 / rate

        self.photon_field = mem_pfield
        return egrid, length


class ContinuousAdiabaticLossRate(object):
    """Implementation of continuous pair production loss rates."""

    def __init__(self, prince_run, energy='grid', *args, **kwargs):
        info(3, 'creating instance')
        #: Reference to species manager
        self.spec_man = prince_run.spec_man

        # Initialize grids
        self.e_cosmicray = prince_run.cr_grid
        self.dim_states = prince_run.dim_states
        self.dim_bins = prince_run.dim_bins
        # Init adiabatic loss vector
        self.energy_vector = self._init_energy_vec(energy)

    def loss_vector(self, z, energy=None):
        """Returns all continuous losses on dim_states grid"""
        # return self.adiabatic_losses(z)
        from prince_cr.cosmology import H
        if energy is None:
            return H(z) * PRINCE_UNITS.cm2sec * self.energy_vector
        else:
            return H(z) * PRINCE_UNITS.cm2sec * energy

    def _init_energy_vec(self, energy):
        """Prepare vector for scaling with units, charge and mass."""
        if energy == 'grid':
            energy_vector = np.zeros(self.dim_states)
            for spec in self.spec_man.species_refs:
                energy_vector[spec.lidx():spec.uidx()] = self.e_cosmicray.grid
        elif energy == 'bins':
            energy_vector = np.zeros(self.dim_bins)
            for spec in self.spec_man.species_refs:
                energy_vector[spec.lbin():spec.ubin()] = self.e_cosmicray.bins
        else:
            raise Exception(
                'Unexpected energy keyword ({:}), use either (grid) or (bins)',
                format(energy))

        return energy_vector

    def single_loss_length(self, pid, z):
        """Returns energy loss length in cm
        (convenience function for plotting)
        """
        species = self.spec_man.ncoid2sref[pid]

        egrid = self.energy_vector[species.sl] * species.A
        rate = self.loss_vector(z)[species.sl] * species.A
        length = egrid / rate
        return egrid, length


class ContinuousPairProductionLossRate(object):
    """Implementation of continuous pair production loss rates."""

    def __init__(self, prince_run, energy='grid', *args, **kwargs):
        info(3, 'creating instance')
        #: Reference to species manager
        self.spec_man = prince_run.spec_man

        #: Reference to PhotonField object
        self.photon_field = prince_run.photon_field

        # Initialize grids
        self.e_cosmicray = prince_run.cr_grid
        self.dim_states = prince_run.dim_states
        self.dim_bins = prince_run.dim_bins
        self.e_photon = prince_run.ph_grid

        # xi is dimensionless (natural units) variable
        xi_steps = 400 if 'xi_steps' not in kwargs else kwargs['xi_steps']
        info(2, 'using', xi_steps, 'steps in xi')
        self.xi = np.logspace(np.log10(2 + 1e-8), 16., xi_steps)

        # weights for integration
        self.phi_xi2 = self._phi(self.xi) / (self.xi**2)

        # Scale vector containing the units and factors of Z**2 for nuclei
        self.scale_vec = self._init_scale_vec(energy)

        # Gamma factor of the cosmic ray
        if energy == 'grid':
            gamma = self.e_cosmicray.grid / PRINCE_UNITS.m_proton
        elif energy == 'bins':
            gamma = self.e_cosmicray.bins / PRINCE_UNITS.m_proton
        else:
            raise Exception(
                'Unexpected energy keyword ({:}), use either (grid) or (bins)',
                format(energy))
        # Grid of photon energies for interpolation
        self.photon_grid = np.outer(1 / gamma,
                                    self.xi) * PRINCE_UNITS.m_electron / 2.
        self.pg_desort = self.photon_grid.reshape(-1).argsort()
        self.pg_sorted = self.photon_grid.reshape(-1)[self.pg_desort]

    def loss_vector(self, z):
        """Returns all continuous losses on dim_states grid"""

        rate_single = trapz(
            self.photon_vector(z) * self.phi_xi2, self.xi, axis=1)
        pprod_loss_vector = self.scale_vec * np.tile(rate_single,
                                                     self.spec_man.nspec)

        return pprod_loss_vector

    def photon_vector(self, z):
        """Returns photon vector at redshift `z` on photon grid.

        This vector is in fact a matrix of vectors of the interpolated
        photon field with dimensions (dim_cr, xi_steps).

        Args:
            z (float): redshift

        Return value from cache if redshift value didn't change since last call.
        """
        photon_vector = np.zeros_like(self.photon_grid)
        photon_vector.reshape(-1)[
            self.pg_desort] = self.photon_field.get_photon_density(
                self.pg_sorted, z)

        return photon_vector

    def _init_scale_vec(self, energy):
        """Prepare vector for scaling with units, charge and mass."""
        if energy == 'grid':
            scale_vec = np.zeros(self.dim_states)
            units = (PRINCE_UNITS.fine_structure * PRINCE_UNITS.r_electron**2 *
                     PRINCE_UNITS.m_electron**2)
            for spec in self.spec_man.species_refs:
                if not spec.is_nucleus:
                    continue
                scale_vec[spec.lidx():spec.uidx()] = units * abs(
                    spec.charge)**2 / float(spec.A) * np.ones_like(
                        self.e_cosmicray.grid, dtype='double')
        elif energy == 'bins':
            scale_vec = np.zeros(self.dim_bins)
            units = (PRINCE_UNITS.fine_structure * PRINCE_UNITS.r_electron**2 *
                     PRINCE_UNITS.m_electron**2)
            for spec in self.spec_man.species_refs:
                if not spec.is_nucleus:
                    continue
                scale_vec[spec.lbin():spec.ubin()] = units * abs(
                    spec.charge)**2 / float(spec.A) * np.ones_like(
                        self.e_cosmicray.bins, dtype='double')
        else:
            raise Exception(
                'Unexpected energy keyword ({:}), use either (grid) or (bins)',
                format(energy))
        return scale_vec

    def single_loss_length(self, pid, z, pfield=None):
        """Returns energy loss length in cm
        (convenience function for plotting)
        """
        if pfield is not None:
            mem_pfield = self.photon_field
            self.photon_field = pfield

        species = self.spec_man.ncoid2sref[pid]

        egrid = self.e_cosmicray.grid * species.A
        rate = self.loss_vector(z)[species.sl] * species.A
        length = egrid / rate

        self.photon_field = mem_pfield
        return egrid, length

    def _phi(self, xi):
        """Phi function as in Blumental 1970"""

        # Simple ultrarelativistic approximation by Blumental 1970
        bltal_ultrarel = np.poly1d([2.667, -14.45, 50.95, -86.07])
        def phi_simple(xi): return xi * bltal_ultrarel(np.log(xi))

        # random fit parameters, see Chorodowski et al
        c1 = 0.8048
        c2 = 0.1459
        c3 = 1.137e-3
        c4 = -3.879e-6

        f1 = 2.91
        f2 = 78.35
        f3 = 1837

        res = np.zeros(xi.shape)

        le = np.where(xi < 25.)
        he = np.where(xi >= 25.)

        res[le] = np.pi / 12. * (xi[le] - 2)**4 / (
            c1 * (xi[le] - 2)**1 + c2 * (xi[le] - 2)**2 + c3 *
            (xi[le] - 2)**3 + c4 * (xi[le] - 2)**4)

        res[he] = phi_simple(
            xi[he]) / (1 - f1 * xi[he]**-1 - f2 * xi[he]**-2 - f3 * xi[he]**-3)

        return res



"""The module contains everything to handle cross section interfaces."""

import numpy as np
from scipy.integrate import trapz

from prince_cr.util import info, get_AZN
from prince_cr.data import spec_data

def get_particle_channels(mo, mo_energy, da_energy):
    """
    Loops over a all daughers for a given mother and generates
    a list of redistribution matrices on the grid:
     np.outer( da_energy , 1 / mo_energy )
    
    Args:
      mo (int): id of the mother particle
      mo_energy (float): energy grid of the mother particle
      da_energy (float): energy grid of the daughter particle (same for all daughters)
    Returns:
      list of np.array: list of redistribution functions on on xgrid 
    """
    info(10, 'Generating decay redistribution for', mo)
    dbentry = spec_data[mo]
    x_grid = np.outer(da_energy, (1 / mo_energy))

    redist = {}
    for branching, daughters in dbentry['branchings']:
        for da in daughters:
            # daughter is a nucleus, we have lorentz factor conservation
            if da > 99:
                res = np.zeros(x_grid.shape)
                res[x_grid == 1.] = 1.
            else:
                res = get_decay_matrix(mo, da, x_grid)
            redist[da] = branching * res

    return x_grid, redist


def get_decay_matrix(mo, da, x_grid):
    """
    Selects the correct redistribution for the given decay channel.
    If the channel is unknown a zero grid is returned instead of raising an error

    Args:
      mo (int): index of the mother
      da (int): index of the daughter
      x_grid (float): grid in x = E_da / E_mo on which to return the result
                      (If x is a 2D matrix only the last column is computed
                      and then repeated over the matrix assuming that the 
                      main diagonal is always x = 1)
    Returns:
      float: redistribution on the grid mo_energy / da_energy
    """

    info(10, 'Generating decay redistribution for', mo, da)

    # --------------------------------
    # pi+ to numu or pi- to nummubar
    # --------------------------------
    if mo in [2, 3] and da in [13, 14]:
        return pion_to_numu(x_grid)

    # --------------------------------
    # pi+ to mu+ or pi- to mu-
    # --------------------------------
    elif mo in [2, 3] and da in [5, 6, 7, 8, 9, 10]:
        # (any helicity)
        if da in [7, 10]:
            return pion_to_muon(x_grid)
        # left handed, hel = -1
        elif da in [5, 8]:
            return pion_to_muon(x_grid) * prob_muon_hel(x_grid, -1.)
        # right handed, hel = 1
        elif da in [6, 9]:
            return pion_to_muon(x_grid) * prob_muon_hel(x_grid, 1.)
        else:
            raise Exception(
                'This should newer have happened, check if-statements above!')

    # --------------------------------
    # muon to neutrino
    # --------------------------------
    elif mo in [5, 6, 7, 8, 9, 10] and da in [11, 12, 13, 14]:
        # translating muon ids to helicity
        muon_hel = {
            5: 1.,
            6: -1.,
            7: 0.,
            8: 1.,
            9: -1.,
            10: 0.,
        }
        hel = muon_hel[mo]
        # muon+ to electron neutrino
        if mo in [5, 6, 7] and da in [11]:
            return muonplus_to_nue(x_grid, hel)
        # muon+ to muon anti-neutrino
        elif mo in [5, 6, 7] and da in [14]:
            return muonplus_to_numubar(x_grid, hel)
        # muon- to elec anti-neutrino
        elif mo in [8, 9, 10] and da in [12]:
            return muonplus_to_nue(x_grid, -1 * hel)
        # muon- to muon neutrino
        elif mo in [8, 9, 10] and da in [13]:
            return muonplus_to_numubar(x_grid, -1 * hel)

    # --------------------------------
    # neutrinos from beta decays
    # --------------------------------

    # beta-
    elif mo > 99 and da == 11:
        info(10, 'nu_e from beta- decay', mo, mo - 1, da)
        return nu_from_beta_decay(x_grid, mo, mo - 1)
    # beta+
    elif mo > 99 and da == 12:
        info(10, 'nubar_e from beta+ decay', mo, mo + 1, da)
        return nu_from_beta_decay(x_grid, mo, mo + 1)
    # neutron
    elif mo > 99 and 99 < da < 200:
        info(10, 'beta decay boost conservation', mo, da)
        return boost_conservation(x_grid)
    else:
        info(
            5,
            'Called with unknown channel {:} to {:}, returning an empty redistribution'.
            format(mo, da))
        # no known channel, return zeros
        return np.zeros(x_grid.shape)


def get_decay_matrix_bin_average(mo, da, x_lower, x_upper):
    """
    Selects the correct redistribution for the given decay channel.
    If the channel is unknown a zero grid is returned instead of raising an error

    Args:
      mo (int): index of the mother
      da (int): index of the daughter
      x_grid (float): grid in x = E_da / E_mo on which to return the result

    Returns:
      float: redistribution on the grid mo_energy / da_energy
    """

    # TODO: Some of the distribution are not averaged yet.
    # The error is small for smooth distributions though
    info(10, 'Generating decay redistribution for', mo, da)

    x_grid = (x_upper + x_lower) / 2

    # remember shape, but only calculate for last column, as x repeats in each column
    from scipy.integrate import trapz
    shape = x_grid.shape

    if len(shape) == 2:
        x_grid = x_grid[:, -1]
        x_upper = x_upper[:, -1]
        x_lower = x_lower[:, -1]

    # --------------------------------
    # pi+ to numu or pi- to nummubar
    # --------------------------------
    if mo in [2, 3] and da in [13, 14]:
        result = pion_to_numu_avg(x_lower, x_upper)

    # --------------------------------
    # pi+ to mu+ or pi- to mu-
    # --------------------------------
    # TODO: The helicity distr need to be averaged analyticaly
    elif mo in [2, 3] and da in [5, 6, 7, 8, 9, 10]:
        # (any helicity)
        if da in [7, 10]:
            result = pion_to_muon_avg(x_lower, x_upper)
        # left handed, hel = -1
        elif da in [5, 8]:
            result = pion_to_muon_avg(x_lower, x_upper) * prob_muon_hel(
                x_grid, -1.)
        # right handed, hel = 1
        elif da in [6, 9]:
            result = pion_to_muon_avg(x_lower, x_upper) * prob_muon_hel(
                x_grid, 1.)
        else:
            raise Exception(
                'This should newer have happened, check if-statements above!')

    # --------------------------------
    # muon to neutrino
    # --------------------------------
    # TODO: The following distr need to be averaged analyticaly
    elif mo in [5, 6, 7, 8, 9, 10] and da in [11, 12, 13, 14]:
        # translating muon ids to helicity
        muon_hel = {
            5: 1.,
            6: -1.,
            7: 0.,
            8: 1.,
            9: -1.,
            10: 0.,
        }
        hel = muon_hel[mo]
        # muon+ to electron neutrino
        if mo in [5, 6, 7] and da in [11]:
            result = muonplus_to_nue(x_grid, hel)
        # muon+ to muon anti-neutrino
        elif mo in [5, 6, 7] and da in [14]:
            result = muonplus_to_numubar(x_grid, hel)
        # muon- to elec anti-neutrino
        elif mo in [8, 9, 10] and da in [12]:
            result = muonplus_to_nue(x_grid, -1 * hel)
        # muon- to muon neutrino
        elif mo in [8, 9, 10] and da in [13]:
            result = muonplus_to_numubar(x_grid, -1 * hel)

    # --------------------------------
    # neutrinos from beta decays
    # --------------------------------
    # TODO: The following beta decay to neutrino distr need to be averaged analyticaly
    # TODO: Also the angular averaging is done numerically still
    # beta-
    elif mo > 99 and da == 11:
        info(10, 'nu_e from beta+ decay', mo, mo - 1, da)
        result = nu_from_beta_decay(x_grid, mo, mo - 1)
    # beta+
    elif mo > 99 and da == 12:
        info(10, 'nubar_e from beta- decay', mo, mo + 1, da)
        result = nu_from_beta_decay(x_grid, mo, mo + 1)
    # neutron
    elif mo > 99 and 99 < da < 200:
        info(10, 'beta decay boost conservation', mo, da)
        result = boost_conservation_avg(x_lower, x_upper)
    else:
        info(
            5,
            'Called with unknown channel {:} to {:}, returning an empty redistribution'.
            format(mo, da))
        # no known channel, return zeros
        result = np.zeros(x_grid.shape)

    # now fill this into diagonals of matrix
    if len(shape) == 2:
        #'filling matrix'
        res_mat = np.zeros(shape)
        for idx, val in enumerate(result[::-1]):
            np.fill_diagonal(res_mat[:, idx:], val)
        result = res_mat

    return result


def pion_to_numu(x):
    """
    Energy distribution of a numu from the decay of pi

    Args:
      x (float): energy fraction transferred to the secondary
    Returns:
      float: probability density at x
    """
    res = np.zeros(x.shape)

    m_muon = spec_data[7]['mass']
    m_pion = spec_data[2]['mass']
    r = m_muon**2 / m_pion**2
    xmin = 0.
    xmax = 1 - r

    cond = np.where(np.logical_and(xmin < x, x <= xmax))
    res[cond] = 1 / (1 - r)
    return res


def pion_to_numu_avg(x_lower, x_upper):
    """
    Energy distribution of a numu from the decay of pi

    Args:
      x_lower,x_lower (float): energy fraction transferred to the secondary, lower/upper bin edge
    Returns:
      float: average probability density in bins (xmin,xmax)
    """
    if x_lower.shape != x_upper.shape:
        raise Exception('different grids for xmin, xmax provided')

    bins_width = x_upper - x_lower
    res = np.zeros(x_lower.shape)

    m_muon = spec_data[7]['mass']
    m_pion = spec_data[2]['mass']
    r = m_muon**2 / m_pion**2
    xmin = 0.
    xmax = 1 - r

    # lower bin edged not contained
    cond = np.where(np.logical_and(xmin > x_lower, xmin < x_upper))
    res[cond] = 1 / (1 - r) * (x_upper[cond] - xmin) / bins_width[cond]

    # upper bin edge not contained
    cond = np.where(np.logical_and(x_lower < xmax, x_upper > xmax))
    res[cond] = 1 / (1 - r) * (xmax - x_lower[cond]) / bins_width[cond]

    # bins fully contained
    cond = np.where(np.logical_and(xmin <= x_lower, x_upper <= xmax))
    res[cond] = 1 / (1 - r)

    return res


def pion_to_muon(x):
    """
    Energy distribution of a muon from the decay of pi

    Args:
      x (float): energy fraction transferred to the secondary
    Returns:
      float: probability density at x
    """
    res = np.zeros(x.shape)

    m_muon = spec_data[7]['mass']
    m_pion = spec_data[2]['mass']
    r = m_muon**2 / m_pion**2
    xmin = r
    xmax = 1.

    cond = np.where(np.logical_and(xmin < x, x <= xmax))
    res[cond] = 1 / (1 - r)
    return res


def pion_to_muon_avg(x_lower, x_upper):
    """
    Energy distribution of a numu from the decay of pi

    Args:
      x_lower,x_lower (float): energy fraction transferred to the secondary, lower/upper bin edge
    Returns:
      float: average probability density in bins (xmin,xmax)
    """
    if x_lower.shape != x_upper.shape:
        raise Exception('different grids for xmin, xmax provided')

    bins_width = x_upper - x_lower
    res = np.zeros(x_lower.shape)

    m_muon = spec_data[7]['mass']
    m_pion = spec_data[2]['mass']
    r = m_muon**2 / m_pion**2
    xmin = r
    xmax = 1.

    # lower bin edged not contained
    cond = np.where(np.logical_and(xmin > x_lower, xmin < x_upper))
    res[cond] = 1 / (1 - r) * (x_upper[cond] - xmin) / bins_width[cond]

    # upper bin edge not contained
    cond = np.where(np.logical_and(x_lower < xmax, x_upper > xmax))
    res[cond] = 1 / (1 - r) * (xmax - x_lower[cond]) / bins_width[cond]

    # bins fully contained
    cond = np.where(np.logical_and(xmin <= x_lower, x_upper <= xmax))
    res[cond] = 1 / (1 - r)

    return res


def prob_muon_hel(x, h):
    """
    Probability for muon+ from pion+ decay to have helicity h
    the result is only valid for x > r

    Args:
      h (int): helicity +/- 1
    Returns:
      float: probability for this helicity
    """

    m_muon = spec_data[7]['mass']
    m_pion = spec_data[2]['mass']

    r = m_muon**2 / m_pion**2

    #helicity expectation value
    hel = 2 * r / (1 - r) / x - (1 + r) / (1 - r)

    res = np.zeros(x.shape)
    cond = np.where(np.logical_and(x > r, x <= 1))
    res[cond] = (1 + hel * h) / 2  #this result is only correct for x > r
    return res


def muonplus_to_numubar(x, h):
    """
    Energy distribution of a numu_bar from the decay of muon+ with helicity h
    (For muon- decay calc with h = -h due to CP invariance)

    Args:
      x (float): energy fraction transferred to the secondary
      h (float): helicity of the muon

    Returns:
      float: probability density at x
    """
    p1 = np.poly1d([4. / 3., -3., 0., 5. / 3.])
    p2 = np.poly1d([-8. / 3., 3., 0., -1. / 3.])

    res = np.zeros(x.shape)
    cond = x <= 1.
    res[cond] = p1(x[cond]) + h * p2(x[cond])
    return res


def muonplus_to_nue(x, h):
    """
    Energy distribution of a n from the decay of muon+ with helicity h
    (For muon- decay calc with h = -h due to CP invariance)

    Args:
      x (float): energy fraction transferred to the secondary
      h (float): helicity of the muon

    Returns:
      float: probability density at x
    """
    p1 = np.poly1d([4., -6., 0., 2.])
    p2 = np.poly1d([-8., 18., -12., 2.])

    res = np.zeros(x.shape)
    cond = x <= 1.
    res[cond] = p1(x[cond]) + h * p2(x[cond])
    return res


def boost_conservation(x):
    """Returns an x=1 distribution for ejected nucleons"""
    dist = np.zeros_like(x)
    # dist[(x == np.max(x)) & (x > 0.9)] = 1.*20.
    dist[x == 1.] = 1. / 0.115
    return dist


def boost_conservation_avg(x_lower, x_upper):
    """Returns an x=1 distribution for ejected nucleons"""
    dist = np.zeros_like(x_lower)

    # boost conservation is a delta peak at x = 1
    # if is is contained in the bin, the the value
    # to 1 / width, else set it to zero
    cond = np.where(np.logical_and(x_lower < 1., x_upper > 1.))
    bins_width = x_upper[cond] - x_lower[cond]
    dist[cond] = 1. / bins_width
    return dist


def nu_from_beta_decay(x_grid, mother, daughter, Gamma=200, angle=None):
    """
    Energy distribution of a neutrinos from beta-decays of mother to daughter
    The res frame distrution is boosted to the observers frame and then angular averaging is done numerically

    Args:
      x_grid (float): energy fraction transferred to the secondary
      mother (int): id of mother
      daughter (int): id of daughter
      Gamma (float): Lorentz factor of the parent particle, default: 200
                     For large Gamma this should not play a role, as the decay is scale invariant
      angle (float): collision angle, if None this will be averaged over 2 pi
    Returns:
      float: probability density on x_grid
    """
    import warnings

    info(10, 'Calculating neutrino energy from beta decay', mother, daughter)

    mass_el = spec_data[20]['mass']
    mass_mo = spec_data[mother]['mass']
    mass_da = spec_data[daughter]['mass']

    Z_mo = spec_data[mother]['charge']
    Z_da = spec_data[daughter]['charge']

    A_mo, _, _ = get_AZN(mother)

    if mother == 100 and daughter == 101:
        # for this channel the masses are already nucleon masses
        qval = mass_mo - mass_da - mass_el
    elif Z_da == Z_mo - 1:  # beta+ decay
        qval = mass_mo - mass_da - 2 * mass_el
    elif Z_da == Z_mo + 1:  # beta- decay
        qval = mass_mo - mass_da
    else:
        raise Exception('Not an allowed beta decay channel: {:} -> {:}'.format(
            mother, daughter))

    # substitute this to the energy grid
    E0 = qval + mass_el
    # NOTE: we subsitute into energy per nucleon here
    Emo = Gamma * mass_mo / A_mo
    E = x_grid * Emo

    # print '------','beta decay','------'
    # print mother
    # print E0
    # print A_mo
    # print Emo

    if angle is None:
        # ctheta = np.linspace(-1, 1, 1000)
        # we use here logspace, as high resolution is mainly needed at small energies
        # otherwise the solution will oscillate at low energy
        ctheta = np.unique(np.concatenate((
                    np.logspace(-8,0,1000) - 1,
                    1 - np.logspace(0,-8,1000),
                )))
    else:
        ctheta = angle

    boost = Gamma * (1 - ctheta)
    Emax = E0 * boost

    E_mesh, boost_mesh = np.meshgrid(E, boost, indexing='ij')
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        res = E_mesh**2 / boost_mesh**5 * (Emax - E_mesh) * np.sqrt(
            (E_mesh - Emax)**2 - boost_mesh**2 * mass_el**2)
    res[E_mesh > Emax] = 0.
    res = np.nan_to_num(res)

    if np.all(res == 0):
        info(10, 'Differential distribution is all zeros for', mother, daughter, 
             'No angle averaging performed!')
    elif angle is None:
        # now average over angle
        res = trapz(res, x=ctheta, axis=1)
        res = res / trapz(res, x=x_grid)
    else:
        res = res[:, 0]
        res = res / trapz(res, x=x_grid)

    return res


def nu_from_beta_decay_old(x_grid, mother, daughter):
    """
    Energy distribution of a neutrinos from beta-decays of mother to daughter

    Args:
      x (float): energy fraction transferred to the secondary
      mother (int): id of mother
      daughter (int): id of daughter
    Returns:
      float: probability density at x
    """

    info(10, 'Calculating neutrino energy from beta decay', mother, daughter)

    mass_el = spec_data[20]['mass']
    mass_mo = spec_data[mother]['mass']
    mass_da = spec_data[daughter]['mass']

    Z_mo = spec_data[mother]['charge']
    Z_da = spec_data[daughter]['charge']

    print(mother, daughter)
    if mother == 100 and daughter == 101:
        # for this channel the masses are already nucleon masses
        qval = mass_mo - mass_da - mass_el
    elif Z_da == Z_mo + 1:  # beta+ decay
        qval = mass_mo - mass_da - 2 * mass_el
    elif Z_da == Z_mo - 1:  # beta- decay
        qval = mass_mo - mass_da
    else:
        raise Exception('Not an allowed beta decay channel: {:} -> {:}'.format(
            mother, daughter))

    E0 = qval + mass_el
    print('Qval', qval, 'E0', E0)
    ye = mass_el / E0
    y_grid = x_grid * mass_mo / 2 / E0

    # norm factor, nomalizing the formula to 1
    norm = 1. / 60. * (np.sqrt(1. - ye**2) * (2 - 9 * ye**2 - 8 * ye**4) +
                       15 * ye**4 * np.log(ye / (1 - np.sqrt(1 - ye**2))))

    cond = y_grid < 1 - ye
    print((1 - ye) * 2 * E0 / mass_mo)
    yshort = y_grid[cond]

    result = np.zeros(y_grid.shape)
    # factor for substitution y -> x
    subst = mass_mo / 2 / E0

    # total formula
    result[cond] = subst / norm * yshort**2 * (1 - yshort) * np.sqrt(
        (1 - yshort)**2 - ye**2)

    result[x_grid > 1] *= 0.

    return result



"""The file provides direct interfaces for matrix multiplications
for Intel MKL runtime library.

Although often, numpy distributions are linked to MKL for the popular
linear algebra ops, sparse algebra always defaults to some basic C++ code
of scipy. Here we explicitly create interfaces to MKL BLAS level 2 and 3
routines.
"""

import numpy as np
from ctypes import c_double, c_int, c_char, POINTER, byref
from prince_cr.config import mkl


def mkl_matdescr():
    """Generate matdescr (matrix description) array required by BLAS calls."""
    from ctypes import c_char, POINTER, byref
    ntrans = byref(c_char(b'n'))  # Non-trans
    trans = byref(c_char(b't'))  # trans
    npmatd = np.chararray(6)
    npmatd[0] = b'G'  # General
    npmatd[3] = b'C'  # C-ordering
    return trans, ntrans, npmatd.ctypes.data_as(POINTER(c_char))


trans, nontrans, matdescr = mkl_matdescr()


def mkl_csrmat_args(csr_mat):
    """Cast pointers of scipy csr matrix for MKL BLAS."""
    data = csr_mat.data.ctypes.data_as(POINTER(c_double))
    ci = csr_mat.indices.ctypes.data_as(POINTER(c_int))
    pb = csr_mat.indptr[:-1].ctypes.data_as(POINTER(c_int))
    pe = csr_mat.indptr[1:].ctypes.data_as(POINTER(c_int))
    m = byref(c_int(csr_mat.shape[0]))
    n = byref(c_int(csr_mat.shape[1]))
    return m, n, data, ci, pb, pe


def mkl_pointer(numpy_array):
    """Cast pointers of numpy dense array of double for MKL."""
    from ctypes import c_double, POINTER
    return numpy_array, numpy_array.ctypes.data_as(POINTER(c_double))


def csrmm(alpha, mA, mB, beta, mC, transa=False):
    """Sparse matrix X dense matrix multiplication.

    Returns C = alpha*A*B+beta*C (in C's memory)
    """
    global trans, nontrans, matdescr
    alpha = byref(c_double(alpha))
    beta = byref(c_double(beta))

    tr = trans if transa else nontrans
    m, k, data, ci, pb, pe = mkl_csrmat_args(mA)
    n = byref(c_int(mC.shape[1]))
    b = mB.ctypes.data_as(POINTER(c_double))
    c = mC.ctypes.data_as(POINTER(c_double))
    ldb = byref(c_int(mB.shape[1]))
    ldc = byref(c_int(mC.shape[1]))
    mkl.mkl_dcsrmm(
        tr, m, n, k, alpha, matdescr,
        data, ci, pb, pe,
        b, ldb,
        beta, c, ldc)
    return mC


def csrmv(alpha, mA, b, beta, c, transa=False):
    """Sparse matrix X dense vector multiplication.

    Returns c = alpha*A*b+beta*c (in C's memory)
    """
    alpha = byref(c_double(alpha))
    beta = byref(c_double(beta))
    tr = trans if transa else nontrans
    m, n, data, ci, pb, pe = mkl_csrmat_args(mA)
    bp = b.ctypes.data_as(POINTER(c_double))
    cp = c.ctypes.data_as(POINTER(c_double))
    mkl.mkl_dcsrmv(
        tr, m, n, alpha, matdescr,
        data, ci, pb, pe,
        bp, beta, cp)
    return c


def dgemv(alpha, mA, vb, beta, vc, transa=False):
    """Dense matrix X dense vector multiplication.

    Declarations not evident from ctypes:
    CblasRowMajor=101, CblasColMajor=102
    CblasNoTrans=111, CblasTrans=112, CblasConjTrans=113

    Returns c = alpha*A*B+beta*c (in c's memory)
    """
    alpha = c_double(alpha)
    beta = c_double(beta)

    # Analyse A layout
    layout = 101 if mA.flags['C_CONTIGUOUS'] else 102
    tr = 112 if transa else 111
    m = mA.shape[0]
    n = mA.shape[1]
    lda = m if layout == 102 else n
    # Increment for the vector (each n-th entry for ex.)
    inc = 1
    a = mA.ctypes.data_as(POINTER(c_double))
    b = vb.ctypes.data_as(POINTER(c_double))
    c = vc.ctypes.data_as(POINTER(c_double))
    mkl.cblas_dgemv(layout, tr, m, n, alpha, a, lda, b,
        inc, beta, c, inc)
    return vc



"""This module contains utility functions, which fulfill common puposes
in different modules of this project."""

import inspect

import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline, RectBivariateSpline
from scipy.integrate import BDF
import prince_cr.config as config

def convert_to_namedtuple(dictionary, name='GenericNamedTuple'):
    """Converts a dictionary to a named tuple."""
    from collections import namedtuple
    return namedtuple(name, list(dictionary.keys()))(**dictionary)

def get_interp_object(xgrid, ygrid, **kwargs):
    """Returns simple standard interpolation object.

    Default type of interpolation is a spline of order
    one without extrapolation (extrapolation to zero).

    Args:
        xgrid (numpy.array): x values of function
        ygrid (numpy.array): y values of function
    """
    if xgrid.shape != ygrid.shape:
        raise Exception(
            'xgrid and ygrid args need identical shapes: {0} != {1}'.format(
                xgrid.shape, ygrid.shape))

    if 'k' not in kwargs:
        kwargs['k'] = 1
    if 'ext' not in kwargs:
        kwargs['ext'] = 'zeros'

    return InterpolatedUnivariateSpline(xgrid, ygrid, **kwargs)
    # if xwidths is not None:
    #     return np.tile(xwidths,len(ygrid)).reshape(len(xwidths),len(ygrid))*res
    # else:
    #     return res


def get_2Dinterp_object(xgrid, ygrid, zgrid, xbins=None, **kwargs):
    """Returns simple standard interpolation object for 2-dimentsional
    distribution.

    Default type of interpolation is a spline of order
    one without extrapolation (extrapolation to zero).

    Args:
        xgrid (numpy.array): x values of function
        ygrid (numpy.array): y values of function
    """
    if (xgrid.shape[0], ygrid.shape[0]) != zgrid.shape:
        raise Exception(
            'x and y grid do not match z grid shape: {0} != {1}'.format(
                (xgrid.shape, ygrid.shape), zgrid.shape))

    if 'kx' not in kwargs:
        kwargs['kx'] = 1
    if 'ky' not in kwargs:
        kwargs['ky'] = 1
    if 's' not in kwargs:
        kwargs['s'] = 0.
    return RectBivariateSplineNoExtrap(xgrid, ygrid, zgrid, xbins, **kwargs)


class RectBivariateSplineNoExtrap(RectBivariateSpline):
    """Same as RectBivariateSpline but makes sure, that extrapolated data is alway 0"""

    def __init__(self, xgrid, ygrid, zgrid, xbins=None, *args, **kwargs):
        self.xbins = xbins
        RectBivariateSpline.__init__(self, xgrid, ygrid, zgrid, *args,
                                     **kwargs)
        xknots, yknots = self.get_knots()
        self.xmin, self.xmax = np.min(xknots), np.max(xknots)
        self.ymin, self.ymax = np.min(yknots), np.max(yknots)

    def __call__(self, x, y, **kwargs):
        if 'grid' not in kwargs:
            x, y = np.meshgrid(x, y)
            kwargs['grid'] = False

            result = RectBivariateSpline.__call__(self, x, y, **kwargs)
            # result = np.where((x < self.xmax) & (x > self.xmin), result, 0.)
            # result[np.isnan(result)] = 0.
            # return result.T
            return np.where(np.isnan(result), 0., result).T
        else:
            result = RectBivariateSpline.__call__(self, x, y, **kwargs)
            # result = np.where((x <= xmax) & (x >= xmin), result, 0.)
            # result[np.isnan(result)] = 0.
            # return result
            return np.where(np.isnan(result), 0., result)


class RectBivariateSplineLogData(RectBivariateSplineNoExtrap):
    """Same as RectBivariateSpline but data is internally interpoled as log(data)"""

    def __init__(self, x, y, z, *args, **kwargs):
        x = np.log10(x)
        y = np.log10(y)

        info(2, 'Spline created')
        RectBivariateSplineNoExtrap.__init__(self, x, y, z, *args, **kwargs)

    def __call__(self, x, y, **kwargs):
        x = np.log10(x)
        y = np.log10(y)

        result = RectBivariateSplineNoExtrap.__call__(self, x, y, **kwargs)
        return result


def caller_name(skip=2):
    """Get a name of a caller in the format module.class.method

    `skip` specifies how many levels of stack to skip while getting caller
    name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.
    An empty string is returned if skipped levels exceed stack height.abs

    From https://gist.github.com/techtonik/2151727
    """
    import inspect

    stack = inspect.stack()
    start = 0 + skip

    if len(stack) < start + 1:
        return ''

    parentframe = stack[start][0]

    name = []

    if config.print_module:
        module = inspect.getmodule(parentframe)
        # `modname` can be None when frame is executed directly in console
        if module:
            name.append(module.__name__ + '.')

    # detect classname
    if 'self' in parentframe.f_locals:
        # I don't know any way to detect call from the object method
        # there seems to be no way to detect static method call - it will
        # be just a function call

        name.append(parentframe.f_locals['self'].__class__.__name__ + '::')

    codename = parentframe.f_code.co_name
    if codename != '<module>':  # top level usually
        name.append(codename + '(): ')  # function or a method
    else:
        name.append(': ')  # If called from module scope

    del parentframe
    return "".join(name)


def info(min_dbg_level, *message, **kwargs):
    """Print to console if `min_debug_level <= config.debug_level`

    The fuction determines automatically the name of caller and appends
    the message to it. Message can be a tuple of strings or objects
    which can be converted to string using `str()`.

    Args:
        min_dbg_level (int): Minimum debug level in config for printing
        message (tuple): Any argument or list of arguments that casts to str
        condition (bool): Print only if condition is True
        blank_caller (bool): blank the caller name (for multiline output)
        no_caller (bool): don't print the name of the caller
    """
    condition = kwargs.pop('condition', min_dbg_level <= config.debug_level)
    # Dont' process the if the function if nothing will happen
    if not (condition or config.override_debug_fcn): 
        return

    blank_caller = kwargs.pop('blank_caller', False)
    no_caller = kwargs.pop('no_caller', False)
    if config.override_debug_fcn and min_dbg_level < config.override_max_level:
        fcn_name = caller_name(skip=2).split('::')[-1].split('():')[0]
        if fcn_name in config.override_debug_fcn:
            min_dbg_level = 0

    if condition and min_dbg_level <= config.debug_level:
        message = [str(m) for m in message]
        cname = caller_name() if not no_caller else ''
        if blank_caller:
            cname = len(cname) * ' '
        print(cname + " ".join(message))

def get_AZN(nco_id):
    """Returns mass number :math:`A`, charge :math:`Z` and neutron
    number :math:`N` of ``nco_id``.

    Args:
        nco_id (int): corsika id of nucleus/mass group
    Returns:
        (int,int,int): (Z,A) tuple
    """
    Z, A = 1, 1

    if nco_id >= 100:
        Z = nco_id % 100
        A = (nco_id - Z) // 100
    else:
        Z, A = 0, 0

    return A, Z, A - Z

def bin_widths(bin_edges):
    """Computes and returns bin widths from given edges."""
    edg = np.array(bin_edges)

    return np.abs(edg[1:, ...] - edg[:-1, ...])

class AdditiveDictionary(dict):
    """This dictionary subclass adds values if keys are
    are already present instead of overwriting. For value tuples
    only the second argument is added and the first kept to its
    original value."""

    def __setitem__(self, key, value):
        if key not in self:
            super(AdditiveDictionary, self).__setitem__(key, value)
        elif isinstance(value, tuple):
            super(AdditiveDictionary, self).__setitem__(
                key, (self[key][0], value[1] + self[key][1]))
        else:
            super(AdditiveDictionary, self).__setitem__(
                key, self[key] + value)
            
class PrinceBDF(BDF):
    """This is a modified version of :class:`scipy.integrate.BDF` solver,
    that avoids oscillations that triggers excessive Jacobian updates.
    This improves solutions for CR propagation for z>1."""
    
    def _step_impl(self):
        from scipy.integrate._ivp.bdf import (
            change_D, solve_bdf_system, NEWTON_MAXITER,
            MIN_FACTOR, MAX_FACTOR, MAX_ORDER)
        from scipy.integrate._ivp.common import norm            
        t = self.t
        D = self.D

        max_step = self.max_step
        min_step = 10 * np.abs(np.nextafter(t, self.direction * np.inf) - t)
        if self.h_abs > max_step:
            h_abs = max_step
            change_D(D, self.order, max_step / self.h_abs)
            self.n_equal_steps = 0
        elif self.h_abs < min_step:
            h_abs = min_step
            change_D(D, self.order, min_step / self.h_abs)
            self.n_equal_steps = 0
        else:
            h_abs = self.h_abs

        atol = self.atol
        rtol = self.rtol
        order = self.order

        alpha = self.alpha
        gamma = self.gamma
        error_const = self.error_const

        J = self.J
        LU = self.LU
        current_jac = self.jac is None

        step_accepted = False
        while not step_accepted:
            if h_abs < min_step:
                return False, self.TOO_SMALL_STEP

            h = h_abs * self.direction
            t_new = t + h

            if self.direction * (t_new - self.t_bound) > 0:
                t_new = self.t_bound
                change_D(D, order, np.abs(t_new - t) / h_abs)
                self.n_equal_steps = 0
                LU = None

            h = t_new - t
            h_abs = np.abs(h)

            y_predict = np.sum(D[:order + 1], axis=0)

            scale = atol + rtol * np.abs(y_predict)
            psi = np.dot(D[1: order + 1].T, gamma[1: order + 1]) / alpha[order]

            converged = False
            c = h / alpha[order]
            while not converged:
                if LU is None:
                    LU = self.lu(self.I - c * J)

                converged, n_iter, y_new, d = solve_bdf_system(
                    self.fun, t_new, y_predict, c, psi, LU, self.solve_lu,
                    scale, self.newton_tol)

                if not converged:
                    if current_jac:
                        break
                    J = self.jac(t_new, y_predict)
                    LU = None
                    current_jac = True

            if not converged:
                factor = 0.5
                h_abs *= factor
                change_D(D, order, factor)
                self.n_equal_steps = 0
                LU = None
                continue

            safety = round(0.9 * (2 * NEWTON_MAXITER + 1) / (2 * NEWTON_MAXITER
                                                       + n_iter), ndigits=15)

            scale = atol + rtol * np.abs(y_new)
            error = error_const[order] * d
            error_norm = norm(error / scale)

            if error_norm > 1:
                factor = max(MIN_FACTOR,
                             safety * error_norm ** (-1 / (order + 1)))
                h_abs *= factor
                change_D(D, order, factor)
                self.n_equal_steps = 0
                # As we didn't have problems with convergence, we don't
                # reset LU here.
            else:
                step_accepted = True

        self.n_equal_steps += 1

        self.t = t_new
        self.y = y_new

        self.h_abs = h_abs
        self.J = J
        self.LU = LU

        # Update differences. The principal relation here is
        # D^{j + 1} y_n = D^{j} y_n - D^{j} y_{n - 1}. Keep in mind that D
        # contained difference for previous interpolating polynomial and
        # d = D^{k + 1} y_n. Thus this elegant code follows.
        D[order + 2] = d - D[order + 1]
        D[order + 1] = d
        for i in reversed(range(order + 1)):
            D[i] += D[i + 1]

        if self.n_equal_steps < order + 1:
            return True, None

        if order > 1:
            error_m = error_const[order - 1] * D[order]
            error_m_norm = norm(error_m / scale)
        else:
            error_m_norm = np.inf
        
        

        if order < MAX_ORDER:
            error_p = error_const[order + 1] * D[order + 2]
            error_p_norm = norm(error_p / scale)
        else:
            error_p_norm = np.inf

        error_norms = np.array([error_m_norm, error_norm, error_p_norm])
        with np.errstate(divide='ignore'): 
            factors = error_norms ** (-1 / np.arange(order, order + 3))

        delta_order = np.argmax(factors) - 1
        order += delta_order
        self.order = order

        factor = min(MAX_FACTOR, safety * np.max(factors))
        
        # # This is the custom modification for PriNCe
        if round(self.h_abs * factor, ndigits=15) > self.max_step:
            if round(self.h_abs,ndigits=15) != self.max_step:
                change_D(D, order, max_step / self.h_abs)
                self.h_abs = self.max_step
                self.n_equal_steps = 0
                self.LU = None
            self.n_equal_steps = 0
            return True, None
        # custom modications end
        self.h_abs *= factor
        change_D(D, order, factor)
        self.n_equal_steps = 0
        self.LU = None

        return True, None

class PrinceProgressBar(object):
    """This is a wrapper around tqdm to process some prince
    argument handling, making it optional, for notebooks and
    python scripts using the bar_type argument."""
    def __init__(self, bar_type=None, nsteps=None):
        if bar_type == None or bar_type == False:
            self.pbar = None
        elif bar_type == 'notebook':
            from tqdm import tqdm_notebook as tqdm
            self.pbar = tqdm(total=nsteps)
            self.pbar.update()
        else:
            from tqdm import tqdm
            self.pbar = tqdm(total=nsteps)
            self.pbar.update()

    def __enter__(self):
        return self
    
    def update(self):
        if self.pbar is not None:
            self.pbar.update()
    
    def __exit__(self, type, value, traceback):
        if self.pbar is not None:
            self.pbar.close()


'''
Created on Feb 22, 2017

@author: Anatoli Fedynitch
'''
from abc import abstractmethod
from os.path import join

import numpy as np
from scipy.integrate import trapz
from scipy.interpolate import UnivariateSpline

import prince_cr.config as config


class PhotonField(object):
    """Base class for constructing target photon densities.

    Derived classes have to implement the method  :func:`PhotonField.get_photon_densities`.
    """

    @abstractmethod
    def get_photon_density(self, E, z):
        raise Exception('Base class method called accidentally.')


class CombinedPhotonField(PhotonField):
    """Class to combine (sum) several models, which inherit from :class:`PhotonField`.

    This class is useful when constructing a realistic photon spectrum, which
    is typically a superposition of CMB and CIB. The list of models can be passed
    to the constructor or each model can be added separately using the :func:`add_model`.

    Args:
          list_of_classes_and_args: Can be either list of classes or list of tuples (class, args)
    """

    def __init__(self, list_of_classes_and_args):

        self.model_list = []

        for arg in list_of_classes_and_args:
            cl, cl_arg = None, None
            if type(arg) not in [tuple, list]:
                cl = arg
                cl_arg = []
            else:
                cl = arg[0]
                cl_arg = arg[1:]

            self.model_list.append(cl(*cl_arg))

    def add_model(self, model_class, model_args=()):
        """Adds a model class to the combination.
        """
        self.model_list.append(model_class(*model_args))

    def get_photon_density(self, E, z):
        """Returns the redshift-scaled number density of photons as a
        sum of different models.

        Args:
          z (float): redshift
          E (float): photon energy (GeV)

        Returns:
          float: CMB photon spectrum in :math:`{\\rm GeV}}^{-1} {\\rm cm}}^{-3}`
        """
        E = np.atleast_1d(E)
        res = np.sum([model.get_photon_density(E, z) for model in self.model_list], axis=0)
        return res


class FlatPhotonSpectrum(PhotonField):
    """Constant photon density for testing.
    """

    def get_photon_density(self, E, z):
        """Returns the redshift-scaled number density of CMB photons

        Args:
          z (float): redshift
          E (float): photon energy (GeV)

        Returns:
          float: CMB photon spectrum in :math:`{\\rm GeV}}^{-1} {\\rm cm}}^{-3}`
        """
        # density at z = 0
        nlocal = np.ones_like(E, dtype='double') * 1e12
        #nlocal = E**-1 * 1e12
        return (1. + z)**2 * nlocal

class CMBPhotonSpectrum(PhotonField):
    """Redshift-scaled number density of CMB photons

    In the CMB frame (equivalent to the observer's frame). Normalisation from Planck's spectrum.
    The scaling goes as :math:`n(E,z) = (1+z)^3 n(E/(1+z), z = 0)`.
    The CMB spectrum is a blackbody spectrum with the present-day temperature T0 = 2.725 K.

    Ref.:
        M. Ahlers, L.A. Anchordoqui, and S. Sarkar, Phys. Rev. D 79, 083009 (2009) [0902.3993]
    """

    def get_photon_density(self, E, z):
        """Returns the redshift-scaled number density of CMB photons

        Args:
          z (float): redshift
          E (float): photon energy (GeV)

        Returns:
          float: CMB photon spectrum in :math:`{\\rm GeV}}^{-1} {\\rm cm}}^{-3}`
        """
        pref = 1.31868e40  # 1/pi^2/(hbar*c)^3 [GeV^-3 cm^-3]
        Ered = E / (1. + z)
        # density at z = 0, for energy E / (1 + z); ECMB = kB * T0

        # Call exp only for values within dynamic range of the function
        eratio = Ered / config.E_CMB
        exp_range = eratio < 709. # This is the limit for 64 bits
        nlocal = np.zeros_like(Ered)
        pref *= (1. + z)**2 # Normalize
        nlocal[exp_range] = pref * Ered[exp_range]**2 / (np.exp(eratio[exp_range]) - 1.0)
        del exp_range
        return nlocal

class EBLSplined2D(PhotonField):
    def __init__(self):
        self.simple_scaling = None
        self.int2d = None

    def get_photon_density(self, E, z):
        """Returns the redshift-scaled number density of CIB photons

        Accepts scalar, vector and matrix arguments.

        Args:
          z (float): redshift
          E (float): photon energy (GeV)

        Returns:
          float: CMB photon spectrum in :math:`{\\rm GeV}}^{-1} {\\rm cm}}^{-3}`
        """
        # pylint:disable=not-callable
        if self.simple_scaling:
            Ered = E / (1. + z)
            nlocal = self.int2d(Ered, 0., assume_sorted=True)
            nz = self.int2d(Ered, z, assume_sorted=True)
            scale = trapz(nz,Ered) / trapz(nlocal,Ered) / (1+z)**3
            # print(scale)
            return (1. + z)**2 * nlocal * scale
        else:
            return self.int2d(E, z, assume_sorted=True)

class CIBFranceschini2D(EBLSplined2D):
    """CIB model "1" by Fraceschini et al.

    CIB photon distribution for z = 0...2. Requires availability of
    an `scipy.interp2d` object file `data/CIB_franceschini_int2D.ppo`.

    Ref.:
        A. Franceschini et al., Astron. Astrphys. 487, 837 (2008) [arXiv:0805.1841]
    """

    def __init__(self, simple_scaling=False):
        from prince_cr.data import db_handler
        self.int2d = db_handler.ebl_spline('Francescini2008','base')
        self.simple_scaling = simple_scaling

class CIBInoue2D(EBLSplined2D):
    """CIB model "2" by Inoue et al.

    CIB photon distribution for z = 0...10. Requires availability of
    an `scipy.interp2d` object file `data/CIB_inoue_int2D.ppo`. A low
    and high variation of the "third-population" component are also
    available, by passing

    Ref.:
        Y. Inoue et al. [arXiv:1212.1683]
    """

    def __init__(self, model='base', simple_scaling=False):
        from prince_cr.data import db_handler

        assert model in ['base', 'upper', 'lower']

        self.int2d = db_handler.ebl_spline('Inoue2013', model)
        self.simple_scaling = simple_scaling
        
class CIBGilmore2D(EBLSplined2D):
    """CIB model "3" by Gilmore et al.

    CIB photon distribution for z = 0...7. Requires availability of
    an `scipy.interp2d` object file `data/CIB_gilmore_int2D.ppo`.

    Note: Currently uses the fixed model from the reference as standard,
          for the fiducial model, change the 'model' keyword

    Ref.:
        R.C. Gilmore et al., MNRAS Soc. 422, 3189 (2012) [arXiv:1104.0671]
    """

    def __init__(self, model='fiducial', simple_scaling=False):
        from prince_cr.data import db_handler

        assert model in ['fixed', 'fiducial']

        self.int2d = db_handler.ebl_spline('Gilmore2011', model)
        self.simple_scaling = simple_scaling


class CIBDominguez2D(EBLSplined2D):
    """CIB model "3" by Gilmore et al.

    CIB photon distribution for z = 0...2. Requires availability of
    an `scipy.interp2d` object file `data/CIB_dominguez_int2D.ppo`.

    Note: The class contains an interpolators for the upper and lower limits,
          which are not yet accessable through a function

    Ref.:
        R.C. Gilmore et al., MNRAS 410, 2556 (2011) [arXiv:1104.0671]
    """
    def __init__(self, model='base', simple_scaling=False):
        from prince_cr.data import db_handler

        assert model in ['base', 'upper', 'lower']

        self.int2d = db_handler.ebl_spline('Dominguez2010', model)
        self.simple_scaling = simple_scaling


class CIBFranceschiniZ0(PhotonField):
    """CIB model "1" by Fraceschini et al.

    CIB photon distribution at z=0.

    Ref.:
        A. Franceschini et al., Astron. Astrphys. 487, 837 (2008) [arXiv:0805.1841]
    """

    def __init__(self):
        self.E = np.array([
            1.00000000e-15, 1.46217717e-12, 2.48885732e-12, 3.54813389e-12,
            4.97737085e-12, 7.31139083e-12, 8.87156012e-12, 1.12979591e-11,
            1.38038426e-11, 1.65958691e-11, 1.77418948e-11, 2.07014135e-11,
            3.10455959e-11, 4.13999675e-11, 4.97737085e-11, 6.22300285e-11,
            8.27942164e-11, 1.03609601e-10, 1.38102010e-10, 1.55417529e-10,
            1.85609414e-10, 2.14387767e-10, 2.75613191e-10, 3.37831390e-10,
            4.17637993e-10, 7.31307454e-10, 9.56400953e-10, 1.24299677e-09,
            1.55417529e-09, 2.25995608e-09, 3.10813590e-09, 6.21584248e-09,
            1.00000000e+01
        ])
        self.ngamma = np.array([
            1.00000000e+06, 7.91224998e+10, 1.78114797e+11, 2.17470323e+11,
            1.87326837e+11, 1.16654098e+11, 7.52488730e+10, 3.74541547e+10,
            1.93508479e+10, 1.01765366e+10, 7.91407205e+09, 5.01880128e+09,
            1.23026877e+09, 5.61047976e+08, 3.23593657e+08, 2.10377844e+08,
            1.08893009e+08, 5.19517283e+07, 2.69649736e+07, 2.26725308e+07,
            2.03422974e+07, 1.70529689e+07, 1.41807761e+07, 1.28617481e+07,
            1.10458718e+07, 5.57057467e+06, 3.36790630e+06, 1.85152679e+06,
            9.67163727e+05, 3.19815862e+05, 9.27897490e+04, 1.09774129e+04,
            1.00000000e-29
        ])

        self.spl_ngamma = UnivariateSpline(self.E, self.ngamma, k=1, s=0)

    def get_photon_density(self, E, z):
        """Returns the redshift-scaled number density of CMB photons

        Args:
          z (float): redshift
          E (float): photon energy (GeV)

        Returns:
          float: CMB photon spectrum in :math:`{\\rm GeV}}^{-1} {\\rm cm}}^{-3}`
        """

        if z > 0:
            raise Exception(self.__class__.__name__ + 'get_photon_density(): '
                            + 'Redshift z > 0 not supported by this class')

        return self.spl_ngamma(E, assume_sorted=True)


class CIBSteckerZ0(PhotonField):
    """CIB model "1" by Stecker et al.

    CIB photon distribution at z=0.

    Ref.:
        F.W. Stecker et al., Astrophys. J. 648, 774 (2006) [astro-ph/0510449]
    """

    def __init__(self):
        self.E = np.array([
            3.30673963e-12, 3.66859694e-12, 4.01143722e-12, 4.32016168e-12,
            4.72389244e-12, 5.16416369e-12, 5.73059874e-12, 6.26469598e-12,
            6.95184371e-12, 7.48858908e-12, 8.06491882e-12, 8.81860723e-12,
            9.49729657e-12, 9.93116048e-12, 1.05390145e-11, 1.13527219e-11,
            1.25979535e-11, 1.43979281e-11, 1.59771753e-11, 1.77296434e-11,
            2.08785339e-11, 2.35125645e-11, 2.81060620e-11, 3.07255737e-11,
            3.46098725e-11, 3.83972393e-11, 4.52168183e-11, 5.24686633e-11,
            6.55692174e-11, 8.19407623e-11, 9.36698878e-11, 1.07075473e-10,
            1.24231005e-10, 1.42010351e-10, 1.67232237e-10, 1.91165713e-10,
            2.05911214e-10, 2.25117885e-10, 2.46116086e-10, 2.81339039e-10,
            3.21595527e-10, 3.46409665e-10, 3.78712814e-10, 4.07934090e-10,
            4.45974472e-10, 4.80374560e-10, 5.17440002e-10, 5.49123462e-10,
            5.82760356e-10, 6.18457699e-10, 6.76129679e-10, 7.28299480e-10,
            7.96214349e-10, 8.32530207e-10, 9.10164726e-10, 9.95061676e-10,
            1.10415489e-09, 1.17179057e-09, 1.26217624e-09, 1.33949162e-09,
            1.42151026e-09, 1.53119323e-09, 1.64930130e-09, 1.85750504e-09,
            2.12334224e-09, 2.28712552e-09, 2.50046051e-09, 2.69333268e-09,
            2.85831442e-09, 3.07878952e-09, 3.31627090e-09, 3.57215257e-09,
            3.79087968e-09, 4.66766845e-09, 5.10305207e-09, 5.57904674e-09,
            6.19070344e-09, 7.07668721e-09, 8.33354913e-09, 1.04145378e-08,
            1.15563317e-08, 1.22642219e-08, 1.28233058e-08, 1.32102185e-08,
            1.40194185e-08, 1.44444164e-08
        ])
        self.ngamma = np.array([
            1.85951634e+11, 1.76116481e+11, 1.69277798e+11, 1.57180983e+11,
            1.43747401e+11, 1.28262588e+11, 1.10001853e+11, 9.33899193e+10,
            7.62254505e+10, 6.40766798e+10, 5.38641735e+10, 4.35211137e+10,
            3.56861749e+10, 3.09100708e+10, 2.57276430e+10, 2.05825889e+10,
            1.48354253e+10, 9.87802225e+09, 6.94592357e+09, 5.00691248e+09,
            2.93042088e+09, 2.13348700e+09, 1.29285554e+09, 1.07082869e+09,
            8.19275561e+08, 6.68574795e+08, 5.01498918e+08, 3.91399260e+08,
            2.63250982e+08, 1.77063890e+08, 1.43783814e+08, 1.13899023e+08,
            8.88934988e+07, 7.21855052e+07, 5.69023235e+07, 4.73685401e+07,
            4.28982919e+07, 3.82771858e+07, 3.41530898e+07, 2.84308602e+07,
            2.36673700e+07, 2.14338408e+07, 1.86560630e+07, 1.68958482e+07,
            1.47061842e+07, 1.29921193e+07, 1.14775713e+07, 1.10868967e+07,
            1.04469616e+07, 9.84396362e+06, 8.78354843e+06, 7.95463030e+06,
            7.09773859e+06, 6.62186008e+06, 5.76381456e+06, 5.01683711e+06,
            4.19681669e+06, 3.76313674e+06, 3.32452804e+06, 2.98098644e+06,
            2.67294486e+06, 2.30350946e+06, 1.93646655e+06, 1.48156258e+06,
            1.11681182e+06, 9.38858192e+05, 7.58577575e+05, 6.22085387e+05,
            5.44126654e+05, 4.35281290e+05, 3.56960367e+05, 3.00088947e+05,
            2.43652011e+05, 1.33045442e+05, 9.97860887e+04, 7.48393488e+04,
            5.39473356e+04, 3.50404762e+04, 1.90370653e+04, 8.19200106e+03,
            5.34724398e+03, 4.23525925e+03, 3.66783677e+03, 3.22403687e+03,
            2.55358313e+03, 2.24460538e+03
        ])

        self.spl_ngamma = UnivariateSpline(self.E, self.ngamma, k=1, s=0)

    def get_photon_density(self, E, z):
        """Returns the redshift-scaled number density of CMB photons

        Args:
          z (float): redshift
          E (float): photon energy (GeV)

        Returns:
          float: CMB photon spectrum in :math:`{\\rm GeV}}^{-1} {\\rm cm}}^{-3}`
        """

        if z > 0:
            raise Exception(self.__class__.__name__ + 'get_photon_density(): '
                            + 'Redshift z > 0 not supported by this class')

        return self.spl_ngamma(E, assume_sorted=True)


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    erange = np.logspace(-20, -6, 100)
    cmb = CMBPhotonSpectrum()
    inoue = CIBInoue2D()
    fig, ax = plt.subplots(1, 1, figsize=(4, 3))
    ax.loglog(
        erange,
        erange * cmb.get_photon_density(erange, z=0.),
        ls='-',
        lw=2,
        color='k')
    ax.set_ylim(1e-9, 1e3)
    #     ax.fill_between(erange, CMB_photon_spectrum(erange, z=0.),
    #                     CMB_photon_spectrum(erange, z=6.), color='b', alpha=0.3)
    ax.set_ylabel(r'$\epsilon$ d$n/$d$\epsilon$ cm$^{-3}$')
    ax.set_xlabel(r'Photon energy $\epsilon$ (GeV)')

    fig, ax = plt.subplots(1, 1, figsize=(4, 3))
    ax.loglog(
        erange,
        erange * inoue.get_photon_density(erange, z=0.),
        ls='-',
        lw=2,
        color='k')
    ax.set_ylabel(r'$\epsilon$ d$n/$d$\epsilon$ cm$^{-3}$')
    ax.set_xlabel(r'Photon energy $\epsilon$ (GeV)')

    mcomb = CombinedPhotonField([CMBPhotonSpectrum, CIBInoue2D])
    fig, ax = plt.subplots(1, 1, figsize=(4, 3))
    ax.loglog(
        erange,
        erange * mcomb.get_photon_density(erange, 0.),
        ls='-',
        lw=2,
        color='k')
    ax.fill_between(
        erange,
        erange * mcomb.get_photon_density(erange, 0.),
        erange * mcomb.get_photon_density(erange, 6.),
        color='r',
        alpha=0.3)
    ax.set_ylabel(r'$\epsilon$ d$n/$d$\epsilon$ cm$^{-3}$')
    ax.set_xlabel(r'Photon energy $\epsilon$ (GeV)')

    plt.show()
